"""
symbolite.impl.libpythoncode
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Utilities to build Python source snippets from Symbolite expressions.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Iterable

from symbolite.core.function import Function, Operator
from symbolite.core.symbolite_object import get_symbolite_info
from symbolite.core.value import Name, Value
from symbolite.ops._get_name import get_full_name


@dataclass(frozen=True)
class CodeExpr:
    """Represents a snippet of Python code plus its precedence."""

    text: str
    precedence: int = 100

    def __str__(self) -> str:  # pragma: no cover - convenience
        return self.text


def _coerce(value: Any) -> CodeExpr:
    if isinstance(value, CodeExpr):
        return value
    if isinstance(value, str):
        return CodeExpr(value)
    if isinstance(value, bool):
        return CodeExpr("True" if value else "False")
    if isinstance(value, (int, float, complex)):
        return CodeExpr(repr(value))
    return CodeExpr(str(value))


def _join_arguments(
    args: Iterable[CodeExpr], kwargs: Iterable[tuple[str, CodeExpr]]
) -> str:
    parts = [arg.text for arg in args]
    parts.extend(f"{key}={value.text}" for key, value in kwargs)
    return ", ".join(parts)


def _maybe_parenthesize(expr: CodeExpr, precedence: int, *, right: bool) -> str:
    if expr.precedence < precedence:
        return f"({expr.text})"
    if right and expr.precedence == precedence:
        return f"({expr.text})"
    return expr.text


def make_function(qualified_name: str) -> Any:
    def _function(*args: Any, **kwargs: Any) -> CodeExpr:
        coerced_args = tuple(_coerce(arg) for arg in args)
        coerced_kwargs = tuple((k, _coerce(v)) for k, v in kwargs.items())
        return CodeExpr(
            f"{qualified_name}({_join_arguments(coerced_args, coerced_kwargs)})"
        )

    return _function


def make_operator(fmt: str, precedence: int, arity: int) -> Any:
    def _operator(*args: Any) -> CodeExpr:
        coerced = tuple(_coerce(arg) for arg in args)
        if arity == 1:
            (value,) = coerced
            formatted = fmt.format(_maybe_parenthesize(value, precedence, right=False))
        else:
            leading = coerced[0]
            trailing = coerced[1:]
            formatted_args = [
                _maybe_parenthesize(leading, precedence, right=False),
                *(
                    _maybe_parenthesize(
                        arg, precedence, right=index == len(trailing) - 1
                    )
                    for index, arg in enumerate(trailing)
                ),
            ]
            formatted = fmt.format(*formatted_args)
        return CodeExpr(formatted, precedence)

    return _operator


def make_attribute(name: str) -> CodeExpr:
    return CodeExpr(name)


def as_function(obj: Function[Any]) -> Any:
    qualified_name = get_full_name(obj)
    return make_function(qualified_name)


def as_operator(obj: Operator[Any]) -> Any:
    info = get_symbolite_info(obj)
    return make_operator(info.fmt, info.precedence, info.arity)


def as_named_value(obj: Value[Any]) -> CodeExpr:
    info = get_symbolite_info(obj)
    if not isinstance(info.value, Name):
        raise ValueError(f"Value {obj!r} is not bound to a Name.")
    qualified_name = get_full_name(obj)
    return make_attribute(qualified_name)



"""
symbolite.impl.libstd.lang
~~~~~~~~~~~~~~~~~~~~~~~~~~

Standard-library implementations for language primitives.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

import types
from typing import Any

import symbolite.impl.libpythoncode as libpythoncode
from symbolite.core import Unsupported

from ...abstract.lang import AssignInfo, BlockInfo
from ...ops._get_name import get_name
from ...ops._translate import translate
from .._lang_value_utils import compile as compile_code

CODE_IMPL = libpythoncode


def Assign(info: AssignInfo, *, libsl: Any) -> Any:
    """Translate AssignInfo into an executable assignment for the target backend."""
    compile_code(translate(info, CODE_IMPL))


def Block(info: BlockInfo, *, libsl: Any) -> Any:
    """Translate a BlockInfo into a callable that executes on the target backend."""
    source = translate(info, CODE_IMPL)
    namespace = compile_code(source, libsl=libsl)
    function = namespace[get_name(info)]
    function.__symbolite_def__ = source
    function.__symbolite_block__ = info
    return function


@translate.register(BlockInfo)
def translate_block_info(obj: BlockInfo, libsl: types.ModuleType) -> Any:
    lang_module = getattr(libsl, "lang", None)
    if lang_module is None or not hasattr(lang_module, "Block"):
        raise Unsupported(
            f"Implementation module {libsl.__name__} does not provide lang.Block translation."
        )
    translate_block = getattr(lang_module, "Block")
    return translate_block(obj, libsl=libsl)


@translate.register(AssignInfo)
def translate_assign_info(obj: AssignInfo, libsl: types.ModuleType) -> Any:
    lang_module = getattr(libsl, "lang", None)
    if lang_module is None or not hasattr(lang_module, "Assign"):
        raise Unsupported(
            f"Implementation module {libsl.__name__} does not provide lang.Assign translation."
        )
    translate_assign = getattr(lang_module, "Assign")
    return translate_assign(obj, libsl=libsl)


def _noop(value: Any, libsl: Any):
    return value


to_bool = _noop
to_int = _noop
to_float = _noop
to_tuple = _noop


def to_list(value: tuple[Any, ...], libsl: Any) -> list[Any]:
    return list(value)


def to_dict(value: tuple[tuple[Any, Any], ...], libsl: Any) -> dict[Any, Any]:
    items = to_tuple(value, libsl)
    return dict(items)


__all__ = [
    "CODE_IMPL",
    "Block",
    "Assign",
    "to_bool",
    "to_int",
    "to_float",
    "to_tuple",
    "to_list",
    "to_dict",
]



"""
symbolite.impl.libstd.vector
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Translate symbolite.abstract.vector
into values and functions defined in Python's math module.

:copyright: 2023 by Symbolite-array Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

import math
import operator

from symbolite.core import Unsupported

getitem = operator.getitem

sum = math.fsum
prod = math.prod

Vector = Unsupported



"""
symbolite.ops._substitute
~~~~~~~~~~~~~~~~~~~~~~~~~

Replace symbols, functions, values, etc by others.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from collections.abc import Mapping
from functools import singledispatch
from typing import Any

from ..core.call import Call
from ..core.symbolite_object import get_symbolite_info
from ..core.value import Name, Value


@singledispatch
def substitute(obj: Any, replacements: Mapping[Any, Any]) -> Any:
    """Replace symbols, functions, values, etc by others.

    Parameters
    ----------
    obj
        symbolic expression.
    replacements
        replacement dictionary.
    """
    return replacements.get(obj, obj)


@substitute.register(Value)
def substitute_value[R: Value[Any]](obj: R, mapper: Mapping[Any, Any]) -> R:
    info = get_symbolite_info(obj)
    if isinstance(info.value, Name):
        return mapper.get(obj, obj)
    return obj.__class__(substitute(info.value, mapper))


@substitute.register
def substitue_call(obj: Call, mapper: Mapping[Any, Any]) -> Call:
    info = get_symbolite_info(obj)
    func = mapper.get(info.func, info.func)
    args = tuple(substitute(arg, mapper) for arg in info.args)
    kwargs = {k: substitute(arg, mapper) for k, arg in info.kwargs_items}

    return Call(func, args, tuple(kwargs.items()))



"""
symbolite.ops._translate
~~~~~~~~~~~~~~~~~~~~~~~~

Translate a symbolic structure using an implementation.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

import types
from functools import singledispatch
from operator import attrgetter
from typing import Any

from ..core import (
    Unsupported,
)
from ..core.call import CallInfo
from ..core.function import FunctionInfo, OperatorInfo, UserFunctionInfo
from ..core.symbolite_object import SymboliteObject, get_symbolite_info
from ..core.value import Name, Value
from ._get_name import get_full_name, get_name, get_namespace


@singledispatch
def translate(obj: Any, libsl: types.ModuleType) -> Any:
    """Translate expression into backend representation.

    Parameters
    ----------
    obj
        symbolic expression.
    libsl
        implementation module.
    """
    return obj


@translate.register
def translate_bool(obj: bool, libsl: types.ModuleType) -> Any:
    return libsl.lang.to_bool(obj, libsl)


@translate.register
def translate_builtin_int(obj: int, libsl: types.ModuleType) -> Any:
    return libsl.lang.to_int(obj, libsl)


@translate.register
def translate_builtin_float(obj: float, libsl: types.ModuleType) -> Any:
    return libsl.lang.to_float(obj, libsl)


@translate.register(tuple)
def translate_tuple(obj: tuple[Any, ...], libsl: types.ModuleType) -> Any:
    return libsl.lang.to_tuple(tuple(translate(value, libsl) for value in obj), libsl)


@translate.register(list)
def translate_list(obj: list[Any], libsl: types.ModuleType) -> Any:
    return libsl.lang.to_list(tuple(translate(value, libsl) for value in obj), libsl)


@translate.register(dict)
def translate_dict(obj: dict[Any, Any], libsl: types.ModuleType) -> Any:
    return libsl.lang.to_dict(
        tuple(
            (translate(key, libsl), translate(value, libsl))
            for key, value in obj.items()
        ),
        libsl,
    )


@translate.register(SymboliteObject)
def translate_symbolite_object(
    obj: SymboliteObject[Any], libsl: types.ModuleType
) -> Any:
    info = get_symbolite_info(obj)
    return translate(info, libsl)


@translate.register(Value)
def translate_value(obj: Value[Any], libsl: types.ModuleType) -> Any:
    info = get_symbolite_info(obj)
    if not isinstance(info.value, Name):
        # Literal value or Call
        return translate(info.value, libsl)

    namespace = get_namespace(info)
    if namespace is None or namespace == "":
        # User defined symbol
        # Try to map the class
        cls_name = f"{obj.__class__.__module__.split('.')[-1]}.{obj.__class__.__name__}"

        cls = attrgetter(cls_name)(libsl)
        if cls is Unsupported:
            raise Unsupported(f"{cls_name} is not supported in module {libsl.__name__}")

        return cls(get_name(obj))
    else:
        # Library symbol
        qname = get_full_name(obj)
        value = attrgetter(qname)(libsl)

        if value is Unsupported:
            raise Unsupported(f"{qname} is not supported in module {libsl.__name__}")

        return value


@translate.register(FunctionInfo)
def translate_function_info(
    obj: FunctionInfo[Any], libsl: types.ModuleType
) -> Any | Unsupported:
    attr_name = f"{obj.namespace}.{obj.name}"
    return attrgetter(attr_name)(libsl)


@translate.register(OperatorInfo)
def translate_operator_info(
    obj: OperatorInfo[Any], libsl: types.ModuleType
) -> Any | Unsupported:
    attr_name = f"{obj.namespace}.{obj.name}"
    return attrgetter(attr_name)(libsl)


@translate.register(CallInfo)
def translate_call_info(obj: CallInfo, libsl: types.ModuleType) -> Any:
    func = translate(obj.func, libsl)
    args = tuple(translate(arg, libsl) for arg in obj.args)
    kwargs = {k: translate(arg, libsl) for k, arg in obj.kwargs_items}

    try:
        return func(*args, **kwargs)
    except Exception as ex:
        try:
            ex.add_note(f"While translating {func}(*{args}, **{kwargs}): {ex}")
        except AttributeError:
            pass
        raise ex


@translate.register(UserFunctionInfo)
def translate_user_function_info(
    obj: UserFunctionInfo[Any, Any, Any], libsl: types.ModuleType
) -> Any:
    impls = obj.impls
    default = None
    for k, v in impls:
        if k is libsl:
            return v
        elif k == "default":
            default = v
    if default is not None:
        return default

    raise Exception(
        f"No implementation found for {libsl.__name__} and no default implementation provided for function {obj!s}"
    )



"""
symbolite.core
~~~~~~~~~~~~~~

Symbolite core classes and functions.

A user should not need to use them directly.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from .call import Call
from .function import Function, Operator
from .value import Value


class Unsupported(ValueError):
    """Label unsupported"""


__all__ = [
    "Call",
    "Value",
    "Function",
    "Operator",
    "Unsupported",
]



"""
symbolite.ops._yield_named
~~~~~~~~~~~~~~~~~~~~~~~~~~

Yields all named structures inside a symbolic structure.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from collections.abc import Generator
from functools import singledispatch
from typing import Any

from ..core import (
    Call,
    Function,
    Value,
)
from ..core.function import Operator, UserFunction
from ..core.symbolite_object import SymboliteObject, get_symbolite_info
from ..core.value import Name


@singledispatch
def yield_named(obj: Any) -> Generator[SymboliteObject[Any], None, None]:
    """Yields all named structures inside a symbolic structure."""
    return
    yield SymboliteObject()  # This is required to make it a generator.


@yield_named.register(tuple)
@yield_named.register(list)
def yield_named_tuple(obj: tuple[Any]) -> Generator[SymboliteObject[Any], None, None]:
    for el in obj:
        yield from yield_named(el)


@yield_named.register(SymboliteObject)
def yield_named_symbolite_object(
    obj: SymboliteObject[Any],
) -> Generator[SymboliteObject[Any], None, None]:
    info = get_symbolite_info(obj)
    yield from yield_named(info)


@yield_named.register(Value)
def yield_named_value(
    obj: Value[Any],
) -> Generator[SymboliteObject[Any], None, None]:
    info = get_symbolite_info(obj)
    if isinstance(info.value, Name):
        yield obj
    else:
        yield from yield_named(info.value)


@yield_named.register(Function | Operator | UserFunction)
def yield_named_base_function(
    obj: Function[Any] | Operator[Any] | UserFunction[Any, Any, Any],
) -> Generator[SymboliteObject[Any], None, None]:
    yield obj


@yield_named.register
def yield_named_call(obj: Call) -> Generator[SymboliteObject[Any], None, None]:
    info = get_symbolite_info(obj)
    yield from yield_named(info.func)
    for arg in info.args:
        yield from yield_named(arg)

    for _, v in info.kwargs_items:
        yield from yield_named(v)



"""
symbolite.impl.libpythoncode.boolean
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Code-emitting counterparts for ``symbolite.abstract.boolean``.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

from symbolite.abstract import boolean as abstract_boolean

from ._codeexpr import CodeExpr, as_operator

and_ = as_operator(abstract_boolean.and_)
xor = as_operator(abstract_boolean.xor)
or_ = as_operator(abstract_boolean.or_)


def Boolean(name: str) -> CodeExpr:
    return CodeExpr(name)


__all__ = ["Boolean", "and_", "xor", "or_"]



import inspect
import types
from typing import Any

import pytest

from symbolite import Symbol, real
from symbolite.abstract.lang import Assign, Block
from symbolite.core.value import Value
from symbolite.impl import get_all_implementations
from symbolite.ops import substitute, translate
from symbolite.ops.base import free_values, value_names

all_impl = get_all_implementations()

x, y, z = map(real.Real, "x y z".split())

xsy = real.Real("xsy")


def _make_block_from_variable(expr: Value[Any], *, name: str = "f") -> Block:
    result = expr.__class__("__result")
    return Block(
        inputs=free_values(expr),
        outputs=(result,),
        lines=(Assign(result, expr),),
        name=name,
    )


def _block_factory_f1() -> Block:
    result = real.Real("result")
    return Block(
        inputs=(x, y),
        outputs=(result,),
        lines=(Assign(result, 2 * x + y),),
        name="f",
    )


def _block_factory_f2() -> Block:
    total = real.Real("total")
    triple = real.Real("triple")
    quadruple = real.Real("quadruple")
    return Block(
        inputs=(x, y, z),
        outputs=(total, triple, quadruple),
        lines=(
            Assign(total, 2 * x + y),
            Assign(triple, 3 * x),
            Assign(quadruple, 4 * z),
        ),
        name="f",
    )


def _block_factory_f3() -> Block:
    total = real.Real("total")
    triple = real.Real("triple")
    quadruple = real.Real("quadruple")
    result: Value[Any] = Value("result")
    return Block(
        inputs=(x, y, z),
        outputs=(result,),
        lines=(
            Assign(total, 2 * x + y),
            Assign(triple, 3 * x),
            Assign(quadruple, 4 * z),
            Assign(
                result,
                {
                    "a": total,
                    "b": triple,
                    "c": quadruple,
                },
            ),
        ),
        name="f",
    )


@pytest.mark.mypy_testing
def test_typing():
    reveal_type(x + y)  # R: symbolite.abstract.real.Real # noqa: F821
    reveal_type(2 + y)  # R: symbolite.abstract.real.Real # noqa: F821
    reveal_type(x + 2)  # R: symbolite.abstract.real.Real # noqa: F821
    reveal_type(x + xsy)  # R: symbolite.abstract.real.Real # noqa: F821
    reveal_type(xsy + x)  # R: symbolite.abstract.real.Real # noqa: F821
    reveal_type(real.cos(x))  # R: symbolite.abstract.real.Real # noqa: F821
    reveal_type(real.cos(xsy))  # R: symbolite.abstract.real.Real # noqa: F821


@pytest.mark.parametrize(
    "expr",
    [
        x + y,
        x - y,
        x * y,
        x / y,
        x**y,
        x // y,
    ],
)
@pytest.mark.parametrize("libsl", all_impl.values(), ids=all_impl.keys())
def test_known_symbols(expr: Value, libsl: types.ModuleType):
    block = _make_block_from_variable(expr)
    f = translate(block, libsl=libsl)
    if inspect.isfunction(f):
        assert f.__name__ == "f"
        assert translate(substitute(expr, {x: 2, y: 3}), libsl=libsl) == f(2, 3)
        assert tuple(inspect.signature(f).parameters.keys()) == ("x", "y")


@pytest.mark.parametrize(
    "block_factory,params,args,result",
    [
        (_block_factory_f1, ("x", "y"), (2, 3), 7),
        (_block_factory_f2, ("x", "y", "z"), (2, 3, 1), (7, 6, 4)),
    ],
)
@pytest.mark.parametrize("libsl", all_impl.values(), ids=all_impl.keys())
def test_block_execution(block_factory, params, args, result, libsl: types.ModuleType):
    block = block_factory()
    f = translate(block, libsl=libsl)
    if inspect.isfunction(f):
        assert f.__name__ == "f"
        assert tuple(inspect.signature(f).parameters.keys()) == params
        assert f(*args) == result


@pytest.mark.parametrize(
    "expr,replaced",
    [
        (x + real.cos(y), 2 + real.cos(3)),
        (x + real.pi * y, 2 + real.pi * 3),
    ],
)
@pytest.mark.parametrize("libsl", all_impl.values(), ids=all_impl.keys())
def test_lib_symbols(expr: Value[Any], replaced: Symbol, libsl: types.ModuleType):
    block = _make_block_from_variable(expr)
    f = translate(block, libsl=libsl)
    if inspect.isfunction(f):
        value = f(2, 3)
        assert f.__name__ == "f"
        assert translate(substitute(expr, {x: 2, y: 3}), libsl=libsl) == value
        assert tuple(inspect.signature(f).parameters.keys()) == ("x", "y")


@pytest.mark.parametrize(
    "expr,namespace,result",
    [
        (
            x + real.pi * real.cos(y),
            None,
            {
                "x",
                "y",
                "real.cos",
                "real.pi",
                "real.mul",
                "real.add",
            },
        ),
        (x + real.pi * real.cos(y), "", {"x", "y"}),
        (
            x + real.pi * real.cos(y),
            "real",
            {"real.cos", "real.pi", "real.mul", "real.add"},
        ),
    ],
)
def test_list_symbols(expr: Value[Any], namespace: str | None, result: Symbol):
    assert value_names(expr, namespace) == result



"""
symbolite.ops._as_code
~~~~~~~~~~~~~~~~~~~~~~

Convert a symbolite object to python code.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from functools import singledispatch
from typing import Any


@singledispatch
def as_code(obj: Any) -> str:
    """Convert a symbolite object to python code."""
    from ..impl import libpythoncode
    from ._translate import translate

    s = translate(obj, libpythoncode)
    if hasattr(s, "text"):
        return s.text
    elif isinstance(s, str):
        return s
    else:
        return str(s)



"""
symbolite.abstract.vector
~~~~~~~~~~~~~~~~~~~~~~~~~

Objects and functions for vector operations.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

from collections.abc import Iterable, Mapping, Sequence
from typing import Any, cast, overload

from ..core import Value
from ..core.function import (
    BinaryOperator,
    UnaryFunction,
    UnaryOperator,
)
from .boolean import Boolean
from .real import NumberT, Real
from .symbol import Symbol

VectorT = Iterable[NumberT]


class Vector(Value[VectorT]):
    """Vector Symbolic value.

    See Symbol and Value for information
    """

    def eq(self, other: Any) -> Boolean:
        return eq(self, other)

    def ne(self, other: Any) -> Boolean:
        return ne(self, other)

    def __getitem__(self, key: int | Real) -> Real:
        return getitem(self, key)

    # Normal arithmetic operators
    def __add__(self, other: Any) -> Vector:
        """Implements addition."""
        return add(self, other)

    def __sub__(self, other: Any) -> Vector:
        """Implements subtraction."""
        return sub(self, other)

    def __mul__(self, other: Any) -> Vector:
        """Implements multiplication."""
        return mul(self, other)

    def __matmul__(self, other: Vector) -> Real:
        """Implements multiplication."""
        return matmul(self, other)

    def __truediv__(self, other: Vector) -> Vector:
        """Implements true division."""
        return truediv(self, other)

    def __floordiv__(self, other: Any) -> Vector:
        """Implements integer division using the // operator."""
        return floordiv(self, other)

    # Reflected arithmetic operators
    def __radd__(self, other: Any) -> Vector:
        """Implements reflected addition."""
        return add(other, self)

    def __rsub__(self, other: Any) -> Vector:
        """Implements reflected subtraction."""
        return sub(other, self)

    def __rmul__(self, other: Any) -> Vector:
        """Implements reflected multiplication."""
        return mul(other, self)

    def __rmatmul__(self, other: Any) -> Real:
        """Implements reflected multiplication."""
        return matmul(other, self)

    def __rtruediv__(self, other: Any) -> Vector:
        """Implements reflected true division."""
        return truediv(other, self)

    def __rfloordiv__(self, other: Any) -> Vector:
        """Implements reflected integer division using the // operator."""
        return floordiv(other, self)

    # Unary operators and functions
    def __neg__(self) -> Vector:
        """Implements behavior for negation (e.g. -some_object)"""
        return neg(self)

    def __pos__(self) -> Vector:
        """Implements behavior for unary positive (e.g. +some_object)"""
        return pos(self)

    def __invert__(self) -> Vector:
        """Implements behavior for inversion using the ~ operator."""
        return invert(self)


CompOp = BinaryOperator[Vector | VectorT, Vector | VectorT, Boolean]

# Comparison methods (not operator)
eq = CompOp("eq", "vector", precedence=-5, fmt="{} == {}", output_type=Boolean)
ne = CompOp("ne", "vector", precedence=-5, fmt="{} != {}", output_type=Boolean)

# Emulating container types
getitem = BinaryOperator[Vector, Real | int, Real](
    "getitem", "vector", precedence=5, fmt="{}[{}]", output_type=Real
)

BinOp = BinaryOperator[
    Vector | VectorT | Real | NumberT, Vector | VectorT | Real | NumberT, Vector
]

# Emulating numeric types
add = BinOp("add", "vector", precedence=0, fmt="{} + {}", output_type=Vector)
sub = BinOp("sub", "vector", precedence=0, fmt="{} - {}", output_type=Vector)
mul = BinOp("mul", "vector", precedence=1, fmt="{} * {}", output_type=Vector)
matmul = BinaryOperator[
    Vector | VectorT | Real | NumberT, Vector | VectorT | Real | NumberT, Real
]("matmul", "vector", precedence=1, fmt="{} @ {}", output_type=Real)
truediv = BinOp("truediv", "vector", precedence=1, fmt="{} / {}", output_type=Vector)
floordiv = BinOp("floordiv", "vector", precedence=1, fmt="{} // {}", output_type=Vector)

UnOp = UnaryOperator[Vector, Vector]

# Unary operatorss
neg = UnOp("neg", "vector", precedence=2, fmt="-{}", output_type=Vector)
pos = UnOp("pos", "vector", precedence=2, fmt="+{}", output_type=Vector)
invert = UnOp("invert", "vector", precedence=2, fmt="~{}", output_type=Vector)

sum = UnaryFunction[Vector, Real]("sum", "vector", output_type=Real)
prod = UnaryFunction[Vector, Real]("prod", "vector", output_type=Real)


@overload
def vectorize[T: Value[Any]](
    expr: NumberT,
    input_names: Sequence[str] | Mapping[str, int],
    output_name: str = "vec",
    scalar_type: type[T] = Real,
) -> NumberT: ...


@overload
def vectorize[T: Value[Any]](
    expr: T,
    input_names: Sequence[str] | Mapping[str, int],
    output_name: str = "vec",
    scalar_type: type[T] = Real,
) -> T: ...


@overload
def vectorize[T: Value[Any]](
    expr: Iterable[NumberT | T],
    input_names: Sequence[str] | Mapping[str, int],
    output_name: str = "vec",
    scalar_type: type[T] = Real,
) -> tuple[NumberT | T, ...]: ...


def vectorize[T: Value[Any]](
    expr: NumberT | T | Iterable[NumberT | T],
    input_names: Sequence[str] | Mapping[str, int],
    output_name: str = "vec",
    scalar_type: type[T] = Real,
) -> NumberT | T | Vector | tuple[NumberT | T | Vector, ...]:
    """Vectorize expression by replacing real values
    by an array at a given indices.

    Parameters
    ----------
    expr
    input_names
        if a tuple, provides the names of the values
        which will be mapped to the indices given by their position.
        if a dict, maps symbol names to indices.
    output_name
        name of the array value
    """

    # It would be nice that NumberT is actually the PT of Value.
    if isinstance(expr, NumberT):
        return expr

    if not isinstance(expr, scalar_type):
        # Then is an iterable, vectorize each element.
        expr = cast(Iterable[NumberT | T], expr)
        return tuple(
            vectorize(symbol, input_names, output_name, scalar_type) for symbol in expr
        )

    if isinstance(input_names, dict):
        it = zip(input_names.values(), input_names.keys())
    else:
        it = enumerate(input_names)

    arr = Vector(output_name)

    reps = {scalar_type(name): arr[ndx] for ndx, name in it}
    from ..ops import substitute

    return substitute(expr, reps)


@overload
def auto_vectorize(
    expr: NumberT,
    output_name: str = "vec",
    scalar_type: type[Real] = Real,
) -> tuple[tuple[str, ...], Real]: ...


@overload
def auto_vectorize(
    expr: Real,
    output_name: str = "vec",
    scalar_type: type[Real] = Real,
) -> tuple[tuple[str, ...], Real]: ...


@overload
def auto_vectorize(
    expr: Iterable[Real],
    output_name: str = "vec",
    scalar_type: type[Real] = Real,
) -> tuple[tuple[str, ...], tuple[Real, ...]]: ...


def auto_vectorize(
    expr: NumberT | Real | Iterable[Real],
    output_name: str = "vec",
    scalar_type: type[Real] = Real,
) -> tuple[
    tuple[str, ...],
    NumberT | Real | tuple[NumberT | Real, ...],
]:
    """Vectorize expression by replacing all test_scalar values
    by an array at a given indices. Symbols are ordered into
    the array alphabetically.

    Parameters
    ----------
    expr
    output_name
        name of the array value

    Returns
    -------
    tuple[str, ...]
        symbol names as ordered in the array.
    SymbolicExpression
        vectorized expression.

    """
    from ..ops.base import value_names as _symbol_names

    if isinstance(expr, NumberT):
        return tuple(), expr

    if not isinstance(expr, (Symbol, Real, Vector)):
        expr = tuple(expr)
        out = set[str]()
        for symbol in expr:
            out.update(_symbol_names(symbol, ""))
        symbol_names = tuple(sorted(out))
        return symbol_names, vectorize(expr, symbol_names, output_name, scalar_type)
    else:
        symbol_names = tuple(sorted(_symbol_names(expr, "")))
        return symbol_names, vectorize(expr, symbol_names, output_name, scalar_type)



"""
symbolite.impl.libsympy.real
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Translate symbolite.abstract.real
into values and functions defined in SymPy.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

import operator as op
from typing import Any

import sympy
import sympy as sy
from sympy.abc import x, y

from symbolite.core import Unsupported

# Comparison methods (not operator)
eq = op.eq
ne = op.ne

# Comparison
lt = op.lt
le = op.le
gt = op.gt
ge = op.ge

# Emulating numeric types
add = op.add
sub = op.sub
mul = op.mul
matmul = op.matmul
truediv = op.truediv
floordiv = op.floordiv
mod = op.mod
pow = op.pow
lshift = op.lshift
rshift = op.rshift
and_ = op.and_
xor = op.xor
or_ = op.or_


def _rev(func: Any) -> Any:
    def _internal(a: Any, b: Any) -> Any:
        return func(b, a)

    return _internal


# Reflective versions
radd = _rev(op.add)
rsub = _rev(op.sub)
rmul = _rev(op.mul)
rmatmul = _rev(op.matmul)
rtruediv = _rev(op.truediv)
rfloordiv = _rev(op.floordiv)
rmod = _rev(op.mod)
rpow = _rev(op.pow)
rlshift = _rev(op.lshift)
rrshift = _rev(op.rshift)
rand = _rev(op.and_)
rxor = _rev(op.xor)
ror = _rev(op.or_)

# Unary operators
neg = op.neg
pos = op.pos
invert = op.inv

abs = sy.Abs
acos = sy.acos
acosh = sy.acosh
asin = sy.asin
asinh = sy.asinh
atan = sy.atan
atan2 = sy.atan2
atanh = sy.atanh
ceil = sy.ceiling
comb = Unsupported
copysign = Unsupported
cos = sy.cos
cosh = sy.cosh
degrees = sy.Lambda(x, x * 180 / sy.pi)
erf = Unsupported
erfc = Unsupported
exp = sy.exp
expm1 = sy.Lambda(x, sy.exp(x) - 1)
fabs = sy.Abs
factorial = Unsupported
floor = sy.floor
fmod = sy.Mod
frexp = Unsupported
gamma = Unsupported
hypot = sy.Lambda((x, y), sy.sqrt(x * x + y * y))
isfinite = Unsupported
isinf = Unsupported
isnan = Unsupported
isqrt = Unsupported
ldexp = Unsupported
lgamma = sy.loggamma
log = sy.log
log10 = sy.Lambda(x, sy.log(x, 10))
log1p = sy.Lambda(x, sy.log(1 + x))
log2 = sy.Lambda(x, sy.log(x, 2))
modf = Unsupported
nextafter = Unsupported
pow = sy.Pow
radians = sy.Lambda(x, x * sy.pi / 180)
remainder = Unsupported
sin = sy.sin
sinh = sy.sinh
sqrt = sy.sqrt
tan = sy.tan
tanh = sy.tanh
trunc = Unsupported
ulp = Unsupported

e = sy.exp(1)
inf = sy.oo
pi = sy.pi
nan = sy.nan
tau = 2 * sy.pi


def Real(name: str):
    return sympy.Symbol(name, real=True)


del sy, Unsupported, x, y



"""
symbolite.impl.libsympy.lang
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

SymPy-backed implementations for language primitives.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

import types
from typing import Any

import sympy
from sympy import Symbol
from sympy.codegen.ast import (
    Assignment,
    CodeBlock,
    FunctionDefinition,
    Return,
    real,
)

from ...abstract.lang import AssignInfo, BlockInfo
from ...ops import get_name, translate


def Assign(info: AssignInfo, libsl: types.ModuleType) -> Assignment:
    """Translate an AssignInfo into a Python assignment statement."""
    return Assignment(Symbol(get_name(info.lhs)), translate(info.rhs, libsl))


def Block(info: BlockInfo, libsl: types.ModuleType):
    """Translate a BlockInfo into a Python function definition."""

    parameters = tuple(Symbol(get_name(var), real=True) for var in info.inputs)

    outputs = tuple(Symbol(get_name(var), real=True) for var in info.outputs)

    if len(outputs) == 0:
        ret = Return()
    elif len(outputs) == 1:
        ret = Return(outputs[0])
    else:
        ret = Return(sympy.Tuple(*outputs))

    return FunctionDefinition(
        return_type=real,
        name=get_name(info),
        parameters=parameters,
        body=CodeBlock(
            *(translate(assign, libsl) for assign in info.lines),
            ret,
        ),
    )


def _return(value: Any, libsl: Any):
    return value


to_bool = _return
to_int = _return
to_float = _return


def to_tuple(value: tuple[Any, ...]):
    return sympy.Tuple(value)


def to_list(value: tuple[Any, ...], libsl: Any) -> list[Any]:
    return list(value)


def to_dict(value: tuple[tuple[Any, Any], ...], libsl: Any) -> dict[Any, Any]:
    items = to_tuple(value, libsl)
    return dict(items)



"""
symbolite.abstract.lang
~~~~~~~~~~~~~~~~~~~~~~~

Language-level symbolic primitives such as assignments and blocks.
"""

from __future__ import annotations

from typing import Any

from symbolite.ops._get_name import get_name

from ..core.lang import AssignInfo, BlockInfo, ValueConverter
from ..core.symbolite_object import (
    SymboliteObject,
    get_symbolite_info,
    set_symbolite_info,
)
from ..core.value import Value
from ..ops._get_name import get_full_name
from ..ops.base import free_values

to_bool = ValueConverter[bool]()

to_int = ValueConverter[int]()

to_float = ValueConverter[float]()

to_tuple = ValueConverter[tuple[Any, ...]]()

to_list = ValueConverter[list[Any]]()

to_dict = ValueConverter[tuple[tuple[Any, Any], ...]]()


def _validate_block_dependencies(info: BlockInfo) -> None:
    defined = {get_full_name(var) for var in info.inputs}

    for line_number, ainfo in enumerate(info.lines, start=1):
        for var in free_values(ainfo.rhs):
            name = get_full_name(var)
            if name not in defined:
                raise ValueError(
                    f"Block line {line_number}: value '{name}' must be provided as an input or defined in a previous line."
                )

        defined.add(get_full_name(ainfo.lhs))

    for output in info.outputs:
        name = get_full_name(output)
        if name not in defined:
            raise ValueError(
                f"Block output value '{name}' must be provided as an input or defined in the block body."
            )


class Assign(SymboliteObject[AssignInfo]):
    def __init__(self, lhs: Value[Any], rhs: Any) -> None:
        set_symbolite_info(self, AssignInfo(lhs, rhs))


class Block(SymboliteObject[BlockInfo]):
    """A block of code with inputs and outputs."""

    def __init__(
        self,
        inputs: tuple[Value[Any], ...],
        outputs: tuple[Value[Any], ...],
        lines: tuple[Assign, ...],
        *,
        name: str = "",
    ) -> None:
        content = tuple(get_symbolite_info(obj) for obj in lines)
        binfo = BlockInfo(inputs, outputs, content, name)
        _validate_block_dependencies(binfo)
        set_symbolite_info(self, binfo)


@get_name.register
def get_name_block_info(obj: BlockInfo, qualified: bool = True) -> str:
    return obj.name or "__symbolite_block"


__all__ = [
    "Assign",
    "AssignInfo",
    "Block",
    "BlockInfo",
    "to_bool",
    "to_int",
    "to_float",
    "to_tuple",
    "to_list",
    "to_dict",
]



"""
symbolite.impl.libpythoncode.vector
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Code-emitting counterparts for ``symbolite.abstract.vector``.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

from symbolite.abstract import vector as abstract_vector

from ._codeexpr import CodeExpr, as_function, as_operator

eq = as_operator(abstract_vector.eq)
ne = as_operator(abstract_vector.ne)

getitem = as_operator(abstract_vector.getitem)

add = as_operator(abstract_vector.add)
sub = as_operator(abstract_vector.sub)
mul = as_operator(abstract_vector.mul)
matmul = as_operator(abstract_vector.matmul)
truediv = as_operator(abstract_vector.truediv)
floordiv = as_operator(abstract_vector.floordiv)

neg = as_operator(abstract_vector.neg)
pos = as_operator(abstract_vector.pos)
invert = as_operator(abstract_vector.invert)

sum = as_function(abstract_vector.sum)
prod = as_function(abstract_vector.prod)


def Vector(name: str) -> CodeExpr:
    return CodeExpr(name)


__all__ = [
    "eq",
    "ne",
    "getitem",
    "add",
    "sub",
    "mul",
    "matmul",
    "truediv",
    "floordiv",
    "neg",
    "pos",
    "invert",
    "sum",
    "prod",
    "Vector",
]



"""
symbolite.impl.libnumpy.vector
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Translate symbolite.abstract.vector
into values and functions defined in NumPy.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

import operator

import numpy as np

from symbolite.core import Unsupported

getitem = operator.getitem

add = np.add

sum = np.sum
prod = np.prod

Vector = Unsupported


del np, operator, Unsupported



"""
symbolite.core.call
~~~~~~~~~~~~~~~~~~~

An call is the result of a function that has been called with
certain arguments.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, NamedTuple

from .symbolite_object import SymboliteObject, set_symbolite_info

if TYPE_CHECKING:
    from .function import SymbolicCallable


class CallInfo(NamedTuple):
    func: SymbolicCallable[Any]
    args: tuple[Any, ...]
    kwargs_items: tuple[tuple[str, Any], ...]


class Call(SymboliteObject[CallInfo]):
    """A Function that has been called with certain arguments."""

    def __init__(
        self,
        func: SymbolicCallable[Any],
        args: tuple[Any, ...],
        kwargs: dict[str, Any] | tuple[tuple[str, Any], ...],
    ) -> None:
        if isinstance(kwargs, dict):
            kwargs = tuple(kwargs.items())

        set_symbolite_info(self, CallInfo(func, args, kwargs))



"""
symbolite.abstract
~~~~~~~~~~~~~~~~~~

Abstract symbolite primitives.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from .boolean import Boolean
from .lang import Assign, Block
from .real import Real
from .symbol import Symbol
from .vector import Vector

__all__ = ["Symbol", "Real", "Vector", "Boolean", "Block", "Assign"]



import pytest

from symbolite import Real, real
from symbolite.ops import as_code

x, y, z = map(real.Real, "x y z".split())


@pytest.mark.parametrize(
    "expr,result",
    [
        (x + y, "x + y"),
        (x - y, "x - y"),
        (x * y, "x * y"),
        (x / y, "x / y"),
        (x**y, "x ** y"),
        (x // y, "x // y"),
        (((x**y) % z), "x ** y % z"),
    ],
)
def test_known_symbols(expr: Real, result: Real):
    assert as_code(expr) == result


@pytest.mark.parametrize(
    "expr,result",
    [
        (x + real.cos(y), "x + real.cos(y)"),
        (x + real.pi, "x + real.pi"),
    ],
)
def test_lib_symbols(expr: Real, result: Real):
    assert as_code(expr) == result



"""
symbolite.ops.base
~~~~~~~~~~~~~~~~~~

Common operations to manipulate symbolic expressions.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

import collections
import types
import warnings
from typing import Any, Callable

from ..core.symbolite_object import get_symbolite_info
from ..core.value import Name, Value
from ._get_name import get_full_name, get_namespace


def count_named(obj: Any) -> dict[Any, int]:
    """Inspect an expression and return what is there
    and how many times.

    Parameters
    ----------
    obj
        symbolic expression.
    """
    from . import yield_named

    cnt = collections.Counter[Any](yield_named(obj))
    if cnt:
        return dict(cnt)
    return {obj: 1}


def evaluate(expr: Any, libsl: types.ModuleType | None = None) -> Any:
    """Translate expression into a backend representation.

    Parameters
    ----------
    expr
        symbolic expression.
    libsl
        implementation module.
    """
    from ..impl import Kind, find_module_in_stack
    from ._translate import translate

    if libsl is None:
        libsl = find_module_in_stack()
    if libsl is None:
        warnings.warn("No libsl provided, defaulting to Python standard library.")
        from ..impl import libstd as libsl
    else:
        if libsl.KIND != Kind.VALUE:
            raise ValueError(
                f"Implementation module {libsl} of kind {libsl.KIND} cannot be used for evaluation."
            )

    return translate(expr, libsl)


def is_free_value(obj: Any) -> bool:
    from ..core.value import Value

    if not isinstance(obj, Value):
        return False

    info = get_symbolite_info(obj)
    return isinstance(info.value, Name) and info.value.namespace == ""


def free_values(obj: Any) -> tuple[Value[Any], ...]:
    from . import yield_named

    seen: list[Any] = []
    for el in filter(is_free_value, yield_named(obj)):
        if el not in seen:
            seen.append(el)

    return tuple(seen)


def compare_namespace(namespace: str) -> Callable[[Any], bool]:
    def _inner(obj: Any) -> bool:
        return get_namespace(obj) == namespace

    return _inner


def value_names(self: Any, namespace: str | None = "") -> set[str]:
    """Return a set of value names (with full namespace indication).

    Parameters
    ----------
    namespace: str or None
        If None, all values will be returned independently of the namespace.
        If a string, will compare valu.namespace to that.
        Defaults to "" which is the namespace for user defined values.
    """
    from . import yield_named

    if namespace is None:
        return set(map(get_full_name, yield_named(self)))
    else:
        return set(
            map(get_full_name, filter(compare_namespace(namespace), yield_named(self)))
        )



"""
symbolite.core.function
~~~~~~~~~~~~~~~~~~~~~~~

Symbolic functions.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

import types
from typing import Any, Callable, Literal, NamedTuple, Protocol, Self, cast

from .call import Call
from .symbolite_object import SymboliteObject, get_symbolite_info, set_symbolite_info
from .value import Value


class SymbolicCallable[O: Value[Any]](Protocol):
    """A callable tha returns a symbolic object."""

    def __call__(self, *args: Any, **kwds: Any) -> O: ...


class FunctionInfo[O: Value[Any]](NamedTuple):
    name: str
    namespace: str
    # Use None to indicate that the arity is unknown or it has keyword arguments
    arity: int | None
    # Result symbolic class
    output_type: type[O]


class Function[O: Value[Any]](SymboliteObject[FunctionInfo[O]]):
    """A callable primitive that will return a Value with a Call as value."""

    def __init__(
        self,
        name: str,
        namespace: str = "",
        *,
        arity: int | None = None,
        output_type: type[O],
    ) -> None:
        set_symbolite_info(self, FunctionInfo(name, namespace, arity, output_type))

    def __call__(self, *args: Any, **kwds: Any) -> O:
        info = get_symbolite_info(self)
        if info.arity is None:
            return info.output_type(Call(self, args, kwds))
        if kwds:
            raise ValueError(
                "If arity is given, keyword arguments should not be provided."
            )
        if len(args) != info.arity:
            raise ValueError(
                f"Invalid number of arguments ({len(args)}), expected {info.arity}."
            )

        expr = Call(self, args, tuple())

        return info.output_type(expr)


class UnaryFunction[I, O: Value[Any]](Function[O]):
    def __init__(self, name: str, namespace: str = "", *, output_type: type[O]) -> None:
        set_symbolite_info(self, FunctionInfo(name, namespace, 1, output_type))

    def __call__(self, arg1: I) -> O:
        return super().__call__(arg1)


class BinaryFunction[I, O: Value[Any]](Function[O]):
    def __init__(self, name: str, namespace: str = "", *, output_type: type[O]) -> None:
        set_symbolite_info(self, FunctionInfo(name, namespace, 2, output_type))

    def __call__(self, arg1: I, arg2: I) -> O:
        return super().__call__(arg1, arg2)


class Function3[I, O: Value[Any]](Function[O]):
    def __init__(self, name: str, namespace: str = "", *, output_type: type[O]) -> None:
        set_symbolite_info(self, FunctionInfo(name, namespace, 3, output_type))

    def __call__(self, arg1: I, arg2: I, arg3: I) -> O:
        return super().__call__(arg1, arg2, arg3)


class OperatorInfo[O: Value[Any]](NamedTuple):
    name: str
    namespace: str
    fmt: str
    arity: int
    output_type: type[O]
    precedence: int


class Operator[O: Value[Any]](SymboliteObject[OperatorInfo[O]]):
    """A callable primitive that will return a Value with a Call as value.

    Unlike Function, Operators has a precedence that allow to write it back
    as string in algebraic manner.
    """

    def __init__(
        self,
        name: str,
        namespace: str = "",
        *,
        fmt: str,
        arity: int,
        output_type: type[O],
        precedence: int = 0,
    ) -> None:
        set_symbolite_info(
            self, OperatorInfo(name, namespace, fmt, arity, output_type, precedence)
        )

    def __call__(self, *args: Any, **kwds: Any) -> O:
        info = get_symbolite_info(self)

        if len(args) != info.arity:
            raise ValueError(
                f"Invalid number of arguments ({len(args)}), expected {info.arity}."
            )

        expr = Call(self, args, tuple())

        return info.output_type(expr)


class UnaryOperator[I, O: Value[Any]](Operator[O]):
    def __init__(
        self,
        name: str,
        namespace: str = "",
        *,
        fmt: str,
        output_type: type[O] = Value,
        precedence: int = 0,
    ) -> None:
        super().__init__(
            name,
            namespace,
            fmt=fmt,
            arity=1,
            output_type=output_type,
            precedence=precedence,
        )

    def __call__(self, arg1: I) -> O:
        return super().__call__(arg1)


class BinaryOperator[I1, I2, O: Value[Any]](Operator[O]):
    def __init__(
        self,
        name: str,
        namespace: str = "",
        *,
        fmt: str,
        output_type: type[O] = Value,
        precedence: int = 0,
    ) -> None:
        super().__init__(
            name,
            namespace,
            fmt=fmt,
            arity=2,
            output_type=output_type,
            precedence=precedence,
        )

    def __call__(self, arg1: I1, arg2: I2) -> O:
        return super().__call__(arg1, arg2)


class UserFunctionInfo[P, T, O: Value[Any]](NamedTuple):
    name: str
    namespace: str
    # Use None to indicate that the arity is unknown or it has keyword arguments
    arity: int | None
    # Result symbolic class
    output_type: type[O]
    impls: tuple[tuple[types.ModuleType | Literal["default"], Callable[[P], T]], ...]


class UserFunction[P, T, O: Value[Any]](SymboliteObject[UserFunctionInfo[P, T, O]]):
    def __init__(
        self,
        name: str,
        namespace: str = "",
        *,
        arity: int | None = None,
        output_type: type[O],
    ) -> None:
        set_symbolite_info(
            self, UserFunctionInfo(name, namespace, arity, output_type, ())
        )

    @classmethod
    def from_function(cls, func: Callable[[P], T]) -> Self:
        obj = cls(
            name=func.__name__,
            namespace="",
            output_type=cast(type[O], Value),
        )
        obj.register_impl(func, libsl="default")
        return obj

    def register_impl(
        self, func: Callable[[P], T], libsl: types.ModuleType | Literal["default"]
    ) -> None:
        info = get_symbolite_info(self)
        info = info._replace(impls=info.impls + ((libsl, func),))
        set_symbolite_info(self, info)

    def __call__(self, *args: Any, **kwds: Any) -> O:
        info = get_symbolite_info(self)
        if info.arity is None:
            return info.output_type(Call(self, args, kwds))
        if kwds:
            raise ValueError(
                "If arity is given, keyword arguments should not be provided."
            )
        if len(args) != info.arity:
            raise ValueError(
                f"Invalid number of arguments ({len(args)}), expected {info.arity}."
            )

        expr = Call(self, args, tuple())

        return info.output_type(expr)



"""
symbolite
~~~~~~~~~

A minimal symbolic python package.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from .abstract import Real, Symbol, Vector, real, vector
from .core.function import UserFunction
from .ops import substitute, translate

__all__ = [
    "Symbol",
    "Real",
    "real",
    "Vector",
    "vector",
    "UserFunction",
    "translate",
    "substitute",
]



"""
symbolite.impl.libsympy.vector
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Translate symbolite.abstract.vector
into values and functions defined in SymPy.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

import operator

import sympy as sy

getitem = operator.getitem

sum = sum
prod = sy.prod

Vector = sy.IndexedBase



"""
symbolite.impl.libnumpy.lang
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

NumPy-backed implementations for language primitives.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

from ..libstd.lang import *  # noqa: F403



import pathlib

PATH = pathlib.Path(__file__).parent / "symbolite" / "testsuite"
GENERATED_PATH = PATH / "generated"
TEST_README = GENERATED_PATH / "test_readme.py"
README = GENERATED_PATH.parent.parent.parent / "README.md"


def setup_test_readme():
    return
    GENERATED_PATH.mkdir(exist_ok=True)

    INDENT = " " * 4
    with TEST_README.open("w") as out, README.open("r") as readme:
        mode = None
        output = []

        output.append("def all_tests():\n")
        for i, line in enumerate(readme.readlines()):
            output.append("\n")
            if mode is None and line.strip() == "```python":
                mode = "first_line"
                output[i] = INDENT + "# line %04d" % i
                # output[i] = 'def test_line_%04d():\n' % i
                continue
            elif line.strip() == "```":
                continue
            elif mode == "first_line":
                if line.strip() == "":
                    mode = None
                    output[i - 1] = "\n"
                    continue
                if line.strip().startswith(">>>"):
                    mode = "doctest"
                    output[i - 2] = (
                        output[i - 1][:-1] + "  " + output[i - 2]
                    )  # move the def line one line up
                    output[i - 1] = '    """\n'
                else:
                    mode = "test"
            if mode in ("doctest", "test"):
                output[i] = "    " + line
            else:
                pass
                # output[i] = '# %s' % line

        output.append(INDENT + '"""\n')
        out.writelines(output)


def pytest_sessionstart(session):
    try:
        import numpy
        import sympy

        setup_test_readme()
        del numpy, sympy
    except ImportError:
        pass


def rm_tree(pth):
    for child in pth.glob("*"):
        if child.is_file():
            child.unlink()
        else:
            rm_tree(child)
    pth.rmdir()


def pytest_sessionfinish(session, exitstatus):
    try:
        rm_tree(GENERATED_PATH)
        pass
    except FileNotFoundError:
        pass



"""
symbolite.impl.libstd.real
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Translate symbolite.abstract.real
into values and functions in Python standard library.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

import math
import operator as op
from typing import Any

from symbolite.core import Unsupported

_pow = pow


# Comparison methods (not operator)
eq = op.eq
ne = op.ne

# Comparison
lt = op.lt
le = op.le
gt = op.gt
ge = op.ge

# Emulating container types
getitem = op.getitem
symgetattr = getattr

# Emulating numeric types
add = op.add
sub = op.sub
mul = op.mul
matmul = op.matmul
truediv = op.truediv
floordiv = op.floordiv
mod = op.mod
pow = op.pow
pow3 = _pow
lshift = op.lshift
rshift = op.rshift
and_ = op.and_
xor = op.xor
or_ = op.or_


def _rev(func: Any) -> Any:
    def _internal(a: Any, b: Any) -> Any:
        return func(b, a)

    return _internal


# Reflective versions
radd = _rev(op.add)
rsub = _rev(op.sub)
rmul = _rev(op.mul)
rmatmul = _rev(op.matmul)
rtruediv = _rev(op.truediv)
rfloordiv = _rev(op.floordiv)
rmod = _rev(op.mod)
rpow = _rev(op.pow)
rlshift = _rev(op.lshift)
rrshift = _rev(op.rshift)
rand = _rev(op.and_)
rxor = _rev(op.xor)
ror = _rev(op.or_)

# Unary operators
neg = op.neg
pos = op.pos
invert = op.inv

# "gcd": None,  # 1 to ---
# "hypot": None,  # 1 to ---
# "isclose": None,  # 2, 3, 4
# "lcm": None,  # 1 to ---
# "perm": None,  # 1 or 2
# "log": None,  # 1 or 2 is used as log(x, e)

abs = abs
acos = math.acos
acosh = math.acosh
asin = math.asin
asinh = math.asinh
atan = math.atan
atan2 = math.atan2
atanh = math.atanh
ceil = math.ceil
comb = math.comb
copysign = math.copysign
cos = math.cos
cosh = math.cosh
degrees = math.degrees
erf = math.erf
erfc = math.erfc
exp = math.exp
expm1 = math.expm1
fabs = math.fabs
factorial = math.factorial
floor = math.floor
fmod = math.fmod
frexp = math.frexp
gamma = math.gamma
hypot = math.hypot
isfinite = math.isfinite
isinf = math.isinf
isnan = math.isnan
isqrt = math.isqrt
ldexp = math.ldexp
lgamma = math.lgamma
log = math.log
log10 = math.log10
log1p = math.log1p
log2 = math.log2
modf = math.modf
nextafter = math.nextafter
log2 = math.log2
# This breaks pint
# pow = math.pow
radians = math.radians
remainder = math.remainder
sin = math.sin
sinh = math.sinh
sqrt = math.sqrt
tan = math.tan
tanh = math.tanh
tan = math.tan
trunc = math.trunc
ulp = math.ulp

e = math.e
inf = math.inf
pi = math.pi
nan = math.nan
tau = math.tau

Real = Unsupported

del math, Unsupported, _rev, op, _pow



import math

import pytest

from symbolite import real
from symbolite.abstract.lang import Assign, Block
from symbolite.core.symbolite_object import get_symbolite_info
from symbolite.impl import libpythoncode, libstd
from symbolite.ops import translate


def _make_sample_block() -> tuple[Block, str]:
    x = real.Real("x")
    y = real.Real("y")
    total = real.Real("total")
    cosine = real.Real("cosine")

    block = Block(
        inputs=(x, y),
        outputs=(total, cosine),
        lines=(
            Assign(total, x + y),
            Assign(cosine, real.cos(total)),
        ),
    )

    expected_definition = (
        "def __symbolite_block(x: real.Real, y: real.Real) -> tuple[real.Real, real.Real]:\n"
        "    total = x + y\n"
        "    cosine = real.cos(total)\n"
        "    return total, cosine"
    )

    return block, expected_definition


def test_block_translate_to_code_definition():
    block, expected_definition = _make_sample_block()

    code = translate(block, libpythoncode)

    assert isinstance(code, str)
    assert code == expected_definition


def test_block_translate_to_callable():
    block, expected_definition = _make_sample_block()

    func = translate(block, libstd)

    assert callable(func)
    assert func.__name__ == "__symbolite_block"
    assert func.__symbolite_def__ == expected_definition
    assert func.__symbolite_block__ == get_symbolite_info(block)

    value = func(2, 3)
    assert isinstance(value, tuple)
    assert value[0] == 5
    assert value[1] == pytest.approx(math.cos(5))


def test_block_requires_defined_user_variables():
    x = real.Real("x")
    total = real.Real("total")
    orphan = real.Real("orphan")

    with pytest.raises(ValueError, match="value 'orphan'"):
        _block = Block(
            inputs=(x,),
            outputs=(total,),
            lines=(Assign(total, x + orphan),),
        )


def test_block_requires_defined_outputs():
    x = real.Real("x")
    total = real.Real("total")

    with pytest.raises(ValueError, match="output value 'total'"):
        _block = Block(
            inputs=(x,),
            outputs=(total,),
            lines=(),
        )


def test_block_custom_name():
    x = real.Real("x")
    y = real.Real("y")
    total = real.Real("total")

    block = Block(
        inputs=(x, y),
        outputs=(total,),
        lines=(Assign(total, x + y),),
        name="add",
    )

    expected_definition = (
        "def add(x: real.Real, y: real.Real) -> real.Real:\n"
        "    total = x + y\n"
        "    return total"
    )

    code = translate(block, libpythoncode)
    assert code == expected_definition

    func = translate(block, libstd)
    assert func.__name__ == "add"
    assert func.__symbolite_def__ == expected_definition
    assert func.__symbolite_block__ == get_symbolite_info(block)


def test_block_assign_translation_uses_backend():
    block, _ = _make_sample_block()

    module = libpythoncode

    original_assign = module.lang.Assign

    def custom_assign(assign, rhs_code=None, **kwargs):
        assert kwargs.get("libsl") is module
        rendered = original_assign(assign, rhs_code=rhs_code, **kwargs)
        return f"{rendered}  # custom"

    module.lang.Assign = custom_assign  # type: ignore[attr-defined]
    try:
        code = translate(block, module)
        assert "# custom" in code
    finally:
        module.lang.Assign = original_assign  # type: ignore[attr-defined]


@pytest.mark.xfail
def test_assign_execution_sets_locals():
    x = real.Real("x")
    y = real.Real("y")
    total = real.Real("total")
    assignment = Assign(total, x + y)

    assert translate(assignment, libstd) == {"total": x + y}


def test_assign_translation_to_code():
    x = real.Real("x")
    y = real.Real("y")
    assignment = Assign(x, y + 1)

    rendered = translate(assignment, libpythoncode)
    assert isinstance(rendered, str)
    assert rendered.startswith("x = ")



import types

import pytest

from symbolite import UserFunction, translate
from symbolite.core.call import Call
from symbolite.core.symbolite_object import get_symbolite_info
from symbolite.impl import get_all_implementations

all_impl = get_all_implementations()


def f_1_1(x):
    return 2 * x


def f_2_1(x, y):
    return x + y


def f_1_2(x):
    return 1 * x, 2 * x


@pytest.mark.parametrize(
    "func,args,result",
    [
        (f_1_1, (1,), 2),
        (f_2_1, (1, 2), 3),
        (f_1_2, (1,), (1, 2)),
    ],
)
@pytest.mark.parametrize("libsl", all_impl.values(), ids=all_impl.keys())
def test_user_functions(func, args, result, libsl: types.ModuleType):
    uf = UserFunction.from_function(func)

    output = uf(*args)

    info = get_symbolite_info(output)
    assert isinstance(info.value, Call)
    value_info = get_symbolite_info(info.value)
    assert value_info.func is uf
    assert value_info.args == args
    assert translate(value_info.func, libsl) is func
    assert translate(output, libsl) == result



"""
symbolite.core.value
~~~~~~~~~~~~~~~~~~~~

Symbolic value.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

import random
import string
from typing import Any, NamedTuple

from .call import Call
from .symbolite_object import (
    SymboliteObject,
    get_symbolite_info,
    set_name,
    set_symbolite_info,
)

PREFIX = "__symbolite_value_"


def id_generator(size: int = 6, chars: str = string.ascii_uppercase + string.digits):
    return "".join(random.choice(chars) for _ in range(size))


class Name(NamedTuple):
    name: str
    namespace: str


class ValueInfo[PT](NamedTuple):
    # Used to specify the actual value (if any) of the value.
    # - Name
    # - Call: the result of a function call or operations.
    # - PT: python types compatible with this value.
    value: Name | Call | PT


class Value[PT](SymboliteObject[ValueInfo[PT]]):
    """A symbolic value."""

    def __init__(self, name_or_value: Call | Name | PT | str = "") -> None:
        if isinstance(name_or_value, str):
            if not name_or_value:
                name_or_value = Name(PREFIX + id_generator(), "")
            else:
                parts = name_or_value.split(".")
                if len(parts) == 1:
                    name_or_value = Name(parts[0], "")
                elif len(parts) == 2:
                    name_or_value = Name(parts[1], parts[0])
                else:
                    raise ValueError(
                        f"Invalid value name: expected <name> or <namespace>.<name>, got {name_or_value!r}"
                    )
        set_symbolite_info(self, ValueInfo(name_or_value))


@set_name.register(Value)
def set_name_value(obj: Value[Any], owner: Any, name: str):
    info = get_symbolite_info(obj)

    value = info.value
    if isinstance(value, Name):
        current_name = value.name

        if value.namespace != "":
            return

        if not current_name.startswith(PREFIX) and current_name != name:
            import warnings

            warnings.warn(
                f"Mismatched names in attribute {name}: {type(obj)} is named {current_name}"
            )

        set_symbolite_info(obj, ValueInfo(Name(name, "")))



"""
symbolite.impl.libpythoncode
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Translate Symbolite expressions into Python source snippets.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from .. import Kind
from . import boolean, lang, real, symbol, vector

KIND = Kind.CODE

__all__ = ["symbol", "real", "vector", "boolean", "lang"]



"""
symbolite.impl.libnumpy.real
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Translate symbolite.abstract.real
into values and functions defined in NumPy.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

import operator as op
from typing import Any

import numpy as np

from symbolite.core import Unsupported

# Comparison methods (not operator)
eq = op.eq
ne = op.ne

# Comparison
lt = op.lt
le = op.le
gt = op.gt
ge = op.ge

# Emulating numeric types
add = op.add
sub = op.sub
mul = op.mul
matmul = op.matmul
truediv = op.truediv
floordiv = op.floordiv
mod = op.mod
pow = op.pow
lshift = op.lshift
rshift = op.rshift
and_ = op.and_
xor = op.xor
or_ = op.or_


def _rev(func: Any) -> Any:
    def _internal(a: Any, b: Any) -> Any:
        return func(b, a)

    return _internal


# Reflective versions
radd = _rev(op.add)
rsub = _rev(op.sub)
rmul = _rev(op.mul)
rmatmul = _rev(op.matmul)
rtruediv = _rev(op.truediv)
rfloordiv = _rev(op.floordiv)
rmod = _rev(op.mod)
rpow = _rev(op.pow)
rlshift = _rev(op.lshift)
rrshift = _rev(op.rshift)
rand = _rev(op.and_)
rxor = _rev(op.xor)
ror = _rev(op.or_)

# Unary operators
neg = op.neg
pos = op.pos
invert = op.inv

abs = np.abs
acos = np.arccos
acosh = np.arccosh
asin = np.arcsin
asinh = np.arcsinh
atan = np.arctan
atan2 = np.arctan2
atanh = np.arctanh
ceil = np.ceil
comb = Unsupported
copysign = np.copysign
cos = np.cos
cosh = np.cosh
degrees = np.degrees
erf = Unsupported
erfc = Unsupported
exp = np.exp
expm1 = np.expm1
fabs = np.fabs
factorial = Unsupported
floor = np.floor
fmod = np.fmod
frexp = np.frexp
gamma = Unsupported
hypot = np.hypot
isfinite = np.isfinite
isinf = np.isinf
isnan = np.isnan
isqrt = Unsupported
ldexp = np.ldexp
lgamma = Unsupported
log = np.log
log10 = np.log10
log1p = np.log1p
log2 = np.log2
modf = np.modf
nextafter = np.nextafter
pow = np.power
radians = np.radians
remainder = np.remainder
sin = np.sin
sinh = np.sinh
sqrt = np.sqrt
tan = np.tan
tanh = np.tanh
trunc = np.trunc
ulp = Unsupported

e = np.e
inf = np.inf
pi = np.pi
nan = np.nan
tau = 2 * np.pi

Real = Unsupported

del np, Unsupported



"""
symbolite.impl.libstd
~~~~~~~~~~~~~~~~~~~~~

Translate symbolite
into values and functions defined in Python standard library.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from .. import Kind
from . import lang, real, symbol, vector

KIND = Kind.VALUE

__all__ = ["symbol", "real", "vector", "lang"]



from typing import Any

import pytest

from symbolite import Symbol
from symbolite.core.call import Call
from symbolite.core.function import UserFunction
from symbolite.core.symbolite_object import get_symbolite_info
from symbolite.impl import find_module_in_stack, libpythoncode
from symbolite.impl.libpythoncode._codeexpr import make_function
from symbolite.ops import substitute
from symbolite.ops._as_code import as_code
from symbolite.ops.base import evaluate, value_names

x, y, z = map(Symbol, "x y z".split())

F = UserFunction("F", output_type=Symbol)  # type: ignore
G = UserFunction("G", output_type=Symbol)  # type: ignore

F.register_impl(make_function("F"), libsl=libpythoncode)  # type: ignore
G.register_impl(make_function("G"), libsl=libpythoncode)  # type: ignore


@pytest.mark.mypy_testing
def test_typing():
    reveal_type(x + y)  # R: symbolite.abstract.symbol.Symbol # noqa: F821
    reveal_type(2 + y)  # R: symbolite.abstract.symbol.Symbol # noqa: F821
    reveal_type(x + 2)  # R: symbolite.abstract.symbol.Symbol # noqa: F821


def _helper_forward_reverse(expr: Symbol) -> tuple[Any, ...]:
    expr_info = get_symbolite_info(expr)
    assert isinstance(expr_info.value, Call)
    call_info = get_symbolite_info(expr_info.value)
    func_info = get_symbolite_info(call_info.func)  # type: ignore
    assert func_info.name == "add"  # type: ignore
    return call_info.args


def test_forward_reverse():
    assert _helper_forward_reverse(x + 1) == (x, 1)
    assert _helper_forward_reverse(1 + x) == (1, x)


@pytest.mark.parametrize(
    "expr,result",
    [
        (x < y, "x < y"),
        (x <= y, "x <= y"),
        (x > y, "x > y"),
        (x >= y, "x >= y"),
        (x[1], "x[1]"),
        (x[z], "x[z]"),
        (x + y, "x + y"),
        (x - y, "x - y"),
        (x * y, "x * y"),
        (x @ y, "x @ y"),
        (x / y, "x / y"),
        (x // y, "x // y"),
        (x % y, "x % y"),
        (x**y, "x ** y"),
        (x**y % z, "x ** y % z"),
        (pow(x, y, z), "symbol.pow3(x, y, z)"),
        (x << y, "x << y"),
        (x >> y, "x >> y"),
        (x & y, "x & y"),
        (x ^ y, "x ^ y"),
        (x | y, "x | y"),
        # Reverse
        (1 + y, "1 + y"),
        (1 - y, "1 - y"),
        (1 * y, "1 * y"),
        (1 @ y, "1 @ y"),
        (1 / y, "1 / y"),
        (1 // y, "1 // y"),
        (1 % y, "1 % y"),
        (1**y, "1 ** y"),
        (1 << y, "1 << y"),
        (1 >> y, "1 >> y"),
        (1 & y, "1 & y"),
        (1 ^ y, "1 ^ y"),
        (1 | y, "1 | y"),
        (-x, "-x"),
        (+x, "+x"),
        (~x, "~x"),
        (F(x), "F(x)"),
        (G(x), "G(x)"),
    ],
)
def test_code(expr: Symbol, result: Symbol):
    assert as_code(expr) == result


@pytest.mark.parametrize(
    "expr,result",
    [
        (x + 2 * y, x + 2 * z),
        (x + 2 * F(y), x + 2 * F(z)),
    ],
)
def test_subs(expr: Symbol, result: Symbol):
    assert substitute(expr, {y: z}) == result


@pytest.mark.parametrize(
    "expr,result",
    [
        (x + y, {"x", "y"}),
        (x[z], {"x", "z"}),
        (F(x), {"F", "x"}),
        (G(x), {"G", "x"}),
    ],
)
def test_symbol_names(expr: Symbol, result: set[str]):
    assert value_names(expr) == result


@pytest.mark.parametrize(
    "expr,result",
    [
        (x + y, {"x", "y", "symbol.add"}),
        (x[z], {"x", "z", "symbol.getitem"}),
        (F(x), {"F", "x"}),
        (G(x), {"x", "G"}),
    ],
)
def test_symbol_names_ops(expr: Symbol, result: set[str]):
    assert value_names(expr, None) == result


@pytest.mark.parametrize(
    "expr,result",
    [
        (x + y, set()),
        (x[z], set()),
        (F(x), set()),
        (
            G(x),
            set(),
        ),
    ],
)
def test_symbol_names_namespace(expr: Symbol, result: Symbol):
    assert value_names(expr, namespace="lib") == result


class Real(Symbol):
    pass


@pytest.mark.parametrize(
    "expr,result",
    [
        (x + 2 * y, 1 + 2 * 3),
        # (x + 2 * F(y), x + 2 * F(z)),
    ],
)
def test_eval_str(expr: Symbol, result: Symbol):
    assert eval(as_code(substitute(expr, {x: 1, y: 3}))) == result


@pytest.mark.parametrize(
    "expr,result",
    [
        (x + 2 * y, 1 + 2 * 3),
        # (x + 2 * F(y), x + 2 * F(z)),
    ],
)
def test_evaluate(expr: Symbol, result: Symbol):
    assert evaluate(substitute(expr, {x: 1, y: 3})) == result


def test_find_libs_in_stack():
    assert find_module_in_stack() is None
    from symbolite.impl import libstd as libsl  # noqa: F401

    assert find_module_in_stack()



"""
symbolite.ops
~~~~~~~~~~~~~

Operations to inspect and manipulate symbolic structures:

- as_code: Convert a symbolite object to python code.
- translate: Translate a symbolite object using a backend module.
- substitue: replac

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from ._as_code import as_code
from ._get_name import get_name, get_namespace
from ._substitute import substitute
from ._translate import translate
from ._tree_view import tree_view
from ._yield_named import yield_named
from .base import count_named

__all__ = [
    "count_named",
    "as_code",
    "get_name",
    "get_namespace",
    "substitute",
    "translate",
    "tree_view",
    "yield_named",
]



"""
symbolite.ops._get_name
~~~~~~~~~~~~~~~~~~~~~~~

Utilities to retrieve symbolic names in a consistent way.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

from functools import singledispatch
from typing import Any

from ..core.function import FunctionInfo, OperatorInfo, UserFunctionInfo
from ..core.symbolite_object import SymboliteObject, get_symbolite_info
from ..core.value import Name, ValueInfo

#################
# get_full_name
#################


@singledispatch
def get_full_name(obj: Name) -> str:
    if obj.namespace:
        return f"{obj.namespace}.{obj.name}"
    return obj.name


@get_full_name.register(ValueInfo)
def get_full_name_value(obj: ValueInfo[Any]) -> str:
    if isinstance(obj.value, Name):
        return get_full_name(obj.value)
    else:
        return "<anonymous>"


@get_full_name.register(SymboliteObject)
def get_full_name_symbolite_object(obj: SymboliteObject[Any]) -> str:
    info = get_symbolite_info(obj)
    return get_full_name(info)


#################
# get_name
#################


@singledispatch
def get_name(obj: Name) -> str:
    return obj.name


@get_name.register(SymboliteObject)
def get_name_symbolite_object(obj: SymboliteObject[Any]) -> str:
    info = get_symbolite_info(obj)
    return get_name(info)


@get_name.register(ValueInfo)
def get_name_value(obj: ValueInfo[Any]) -> str:
    if isinstance(obj.value, Name):
        return get_name(obj.value)
    else:
        return "<anonymous>"


@get_name.register(FunctionInfo)
def get_name_function(obj: FunctionInfo[Any]) -> str:
    if obj.namespace:
        return f"{obj.namespace}.{obj.name}"
    return obj.name


@get_name.register(OperatorInfo)
def get_name_operator(obj: OperatorInfo[Any]) -> str:
    if obj.namespace:
        return f"{obj.namespace}.{obj.name}"
    return obj.name


@get_name.register(UserFunctionInfo)
def get_name_user_function_info(obj: UserFunctionInfo[Any, Any, Any]) -> str:
    if obj.namespace:
        return f"{obj.namespace}.{obj.name}"
    return obj.name


#################
# get_namespace
#################


@singledispatch
def get_namespace(obj: Any) -> str | None:
    return None


@get_namespace.register(SymboliteObject)
def get_namespace_symbolite_object(obj: SymboliteObject[Any]) -> str | None:
    info = get_symbolite_info(obj)
    return get_namespace(info)


@get_namespace.register(ValueInfo)
def get_namespace_value(obj: ValueInfo[Any]) -> str | None:
    if isinstance(obj.value, Name):
        return obj.value.namespace
    return None


@get_namespace.register(FunctionInfo)
def get_namespace_function(obj: FunctionInfo[Any]) -> str | None:
    return obj.namespace


@get_namespace.register(OperatorInfo)
def get_namespace_operator(obj: OperatorInfo[Any]) -> str | None:
    return obj.namespace


@get_namespace.register(UserFunctionInfo)
def get_namespace_user_function(obj: UserFunctionInfo[Any, Any, Any]) -> str | None:
    return obj.namespace



"""
symbolite.impl._lang_value_utils
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Runtime helpers for language-level constructs.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

import types
import warnings
from typing import Any

from . import find_module_in_stack


def compile(
    code: str,
    libsl: types.ModuleType | None = None,
) -> dict[str, Any]:
    """Compile code for a given implementation module and return the namespace."""

    if libsl is None:
        libsl = find_module_in_stack()
    if libsl is None:
        warnings.warn("No libsl provided, defaulting to Python standard library.")
        from . import libstd as libsl

    assert libsl is not None

    namespace: dict[str, Any] = {}
    exec(
        code,
        {
            "symbol": libsl.symbol,
            "real": libsl.real,
            # "bool": libsl.boolean, # TODO
            "vector": libsl.vector,
            **globals(),
        },
        namespace,
    )
    return namespace


__all__ = ["compile"]



"""
symbolite.ops._tree_view
~~~~~~~~~~~~~~~~~~~~~~~~

Visualize symbolic structures.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

import itertools
import sys
from functools import singledispatch
from typing import Any, Protocol, TypeVar

from ..core import Value
from ..core.call import CallInfo
from ..core.function import FunctionInfo, OperatorInfo
from ..core.symbolite_object import SymboliteObject, get_symbolite_info
from ..core.value import Name
from ..ops._get_name import get_full_name

_T_contra = TypeVar("_T_contra", contravariant=True)


class SupportsWrite(Protocol[_T_contra]):
    def write(self, s: _T_contra, /) -> object: ...


class Printer:
    file: SupportsWrite[str]
    indent_level: int = 0
    indent_width: int
    _current_line: list[str]

    def __init__(
        self, file: SupportsWrite[str] | None = None, indent_width: int = 2
    ) -> None:
        self.file = file or sys.stdout
        self.indent_width = indent_width
        self._current_line = []

    def append(self, value: str):
        self._current_line.append(value)

    def flush_line(self):
        self.file.write(
            " " * self.indent_width * self.indent_level
            + "".join(self._current_line)
            + "\n"
        )
        self._current_line = []

    def indent(self):
        self.flush_line()
        self.indent_level += 1

    def dedent(self):
        self.flush_line()
        self.indent_level -= 1

    def __del__(self):
        self.flush_line()


def _default_printer() -> Printer:
    return Printer()


@singledispatch
def tree_view(obj: Any, pretty_printer: Printer | None = None):
    pretty_printer = pretty_printer or _default_printer()
    pretty_printer.append(str(obj))


@tree_view.register(SymboliteObject)
def tree_view_symbolite_object(
    obj: SymboliteObject[Any], pretty_printer: Printer | None = None
):
    pretty_printer = pretty_printer or _default_printer()
    info = get_symbolite_info(obj)
    tree_view(info, pretty_printer)


@tree_view.register(FunctionInfo | OperatorInfo)
def tree_view_function(
    obj: FunctionInfo[Any] | OperatorInfo[Any], pretty_printer: Printer | None = None
):
    pretty_printer = pretty_printer or _default_printer()
    pretty_printer.append(f"{obj.namespace}.{obj.name}")


@tree_view.register
def tree_view_call(obj: CallInfo, pretty_printer: Printer | None = None):
    pretty_printer = pretty_printer or _default_printer()
    tree_view(obj.func, pretty_printer)
    pretty_printer.append("(")
    pretty_printer.indent()

    for ndx, (k, v) in enumerate(
        itertools.chain(map(lambda arg: (None, arg), obj.args), obj.kwargs_items)
    ):
        if k is not None:
            pretty_printer.append(f"{k} = ")
        tree_view(v, pretty_printer)
        if ndx < len(obj.args) + len(obj.kwargs_items) - 1:
            pretty_printer.append(",")
            pretty_printer.flush_line()
    pretty_printer.dedent()
    pretty_printer.append(")")


@tree_view.register(Value)
def tree_view_value(obj: Value[Any], pretty_printer: Printer | None = None):
    pretty_printer = pretty_printer or _default_printer()
    info = get_symbolite_info(obj)
    if isinstance(info.value, Name):
        pretty_printer.append(get_full_name(info.value))
    else:
        pretty_printer.append(f"{obj.__class__.__name__}#")
        tree_view(info.value, pretty_printer)



"""
symbolite.core.symbolite_object
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Provides basic funtionality for all symbolic objects.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

from functools import singledispatch
from typing import Any, NamedTuple, Self, cast


class SymboliteObject[R: NamedTuple]:
    """Base class for all Symbolite objects.

    Symbolite information is stored in the __symbolite_info__ attributed
    using a NamedTuple specific for each type of symbolic object.
    In this way we make objects opaque
    """

    __symbolite_info__: R

    def __eq__(self, other: object) -> bool:
        if self.__class__ is not other.__class__:
            return False
        other_named = cast(Self, other)
        return get_symbolite_info(self) == get_symbolite_info(other_named)

    def __hash__(self) -> int:
        return hash(get_symbolite_info(self))

    def __str__(self) -> str:
        return self.__class__.__name__ + "#" + str(get_symbolite_info(self))

    def __repr__(self) -> str:
        return self.__class__.__name__ + "#" + repr(get_symbolite_info(self))

    def __set_name__(self, owner: Any, name: str):
        if name.endswith("__return"):
            return

        set_name(self, owner, name)


def get_symbolite_info[R: NamedTuple](obj: SymboliteObject[R]) -> R:
    return obj.__symbolite_info__


def set_symbolite_info[R: NamedTuple](obj: SymboliteObject[R], info: R):
    object.__setattr__(obj, "__symbolite_info__", info)


@singledispatch
def set_name(obj: Any, owner: Any, name: str):
    """Sets the name of a symbolic object when assigned as an attribute.

    This is called automatically by Python when an object is assigned
    as an attribute of a class or instance.
    """
    return



"""
symbolite.impl.libnumpy.symbol
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Translate symbolite.abstract.symbol
into values and functions defined in NumPy.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

from ..libstd.symbol import *  # noqa: F403



"""
symbolite.impl.libpythoncode.lang
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Code-emitting counterparts for language primitives.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

from typing import Any

from typing_extensions import type_repr

import symbolite.impl.libpythoncode as libpythoncode

from ...abstract.lang import AssignInfo, BlockInfo
from ...ops._get_name import get_full_name, get_name
from ...ops._translate import translate


def Block(info: BlockInfo, **_: Any) -> str:
    """Translate a BlockInfo into a Python function definition."""

    parameters = tuple(
        f"{get_name(var)}: {type_repr(var.__class__).removeprefix('symbolite.abstract.')}"
        for var in info.inputs
    )
    body = [translate(assign, libpythoncode) for assign in info.lines]

    ret = {
        f"{get_full_name(var)}": f"{type_repr(var.__class__).removeprefix('symbolite.abstract.')}"
        for var in info.outputs
    }

    match len(info.outputs):
        case 0:
            return_vars, return_ann = "None", "None"
        case 1:
            return_vars, return_ann = ret.popitem()
        case _:
            # Note: this assumes iter keys an values separately preserves order
            return_vars = ", ".join(ret.keys())
            return_ann = "tuple[" + ", ".join(ret.values()) + "]"

    header = f"def {get_name(info)}({', '.join(parameters)}) -> {return_ann}"

    parts = [f"{header}:"]
    if body:
        parts.append("    " + "\n    ".join(body))
    parts.append(f"    return {return_vars}")

    return "\n".join(parts)


def Assign(info: AssignInfo, **_: Any) -> str:
    """Translate an AssignInfo into a Python assignment statement."""

    lhs = get_name(info.lhs)
    rhs = translate(info.rhs, libpythoncode)
    return f"{lhs} = {rhs}"


def to_bool(value: bool, libsl: Any) -> str:
    return "True" if value else "False"


def to_int(value: int, libsl: Any) -> str:
    return repr(value)


def to_float(value: float, libsl: Any) -> str:
    return repr(value)


def to_tuple(value: tuple[Any, ...], libsl: Any) -> str:
    value = (translate(v, libsl) for v in value)
    return f"({', '.join(value)}, )"


def to_list(value: tuple[Any, ...], libsl: Any) -> str:
    value = (translate(v, libsl) for v in value)
    return f"[{', '.join(value)}]"


def to_dict(value: tuple[tuple[Any, Any], ...], libsl: Any) -> str:
    items = to_tuple(value, libsl)
    return f"dict({items})"


__all__ = [
    "Block",
    "Assign",
    "to_bool",
    "to_int",
    "to_float",
    "to_tuple",
    "to_list",
    "to_dict",
]



import pickle
from typing import Any

from pytest import mark

from .. import Real, Symbol, Vector


@mark.parametrize("cls", [Real, Symbol, Vector])
def test_pickle(cls: Any):
    obj = cls("x")
    dump = pickle.dumps(obj)
    load = pickle.loads(dump)
    assert load == obj



"""
symbolite.impl.libstd.symbol
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Translate symbolite.abstract.symbol
into values and functions in Python standard library.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

import operator as op
import typing as ty

_pow = pow


# Comparison methods (not operator)
eq = op.eq
ne = op.ne

# Comparison
lt = op.lt
le = op.le
gt = op.gt
ge = op.ge

# Emulating container types
getitem = op.getitem
symgetattr = getattr

# Emulating numeric types
add = op.add
sub = op.sub
mul = op.mul
matmul = op.matmul
truediv = op.truediv
floordiv = op.floordiv
mod = op.mod
pow = op.pow
pow3 = _pow
lshift = op.lshift
rshift = op.rshift
and_ = op.and_
xor = op.xor
or_ = op.or_


def _rev(func: ty.Any) -> ty.Any:
    def _internal(a: ty.Any, b: ty.Any) -> ty.Any:
        return func(b, a)

    return _internal


# Reflective versions
radd = _rev(op.add)
rsub = _rev(op.sub)
rmul = _rev(op.mul)
rmatmul = _rev(op.matmul)
rtruediv = _rev(op.truediv)
rfloordiv = _rev(op.floordiv)
rmod = _rev(op.mod)
rpow = _rev(op.pow)
rlshift = _rev(op.lshift)
rrshift = _rev(op.rshift)
rand = _rev(op.and_)
rxor = _rev(op.xor)
ror = _rev(op.or_)

# Unary operators
neg = op.neg
pos = op.pos
invert = op.inv


del _rev, op, _pow



"""
symbolite.core.lang
~~~~~~~~~~~~~~~~~~~

Language related primitives.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

import types
from typing import Any, NamedTuple

from .value import Value


class ValueConverter[T]:
    def __call__(self, value: T, libsl: types.ModuleType) -> Any:
        return value


class AssignInfo(NamedTuple):
    lhs: Value[Any]
    rhs: Any


class BlockInfo(NamedTuple):
    inputs: tuple[Value[Any], ...]
    outputs: tuple[Value[Any], ...]
    lines: tuple[AssignInfo, ...]
    name: str = ""



"""
symbolite.abstract.symbol
~~~~~~~~~~~~~~~~~~~~~~~~~

Objects and functions for symbol operations.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

from typing import Any

from ..core import Value
from ..core.function import (
    BinaryOperator,
    Function3,
    UnaryOperator,
)
from .boolean import Boolean


class Symbol(Value[Any]):
    """General purpose symbolic value.

    The following magic methods are not mapped to symbolite Functions
      - __hash__, __eq__, __ne__ collides with reasonble use of comparisons
        within user code (including uses as dict keys).
        We defined `.eq` y `.ne` methods for the two lasts.
      - __contains__ is coerced to boolean.
      - __bool__ yields a TypeError if not boolean.
      - __str__, __bytes__, __repr__ yields a TypeError if the return value
        is not of the corresponding type.
        and they might also affect usability in the console.
      - __format__
      - __int__, __float__, __complex__ yields a TypeError if the return value
        is not of the corresponding type.
      - __round__, __abs__, __divmod__ they are too "numeric related"
      - __trunc__, __ceil__, __floor__ they are too "numeric related"
        and called by functions in math.
      - __len__ yields a TypeError if not int.
      - __index__ yields a TypeError if not int.

    Also, magic methods that are statements (not expressions) are also not
    mapped: e.g. __setitem__ or __delitem__

    See Value for more information.
    """

    # Comparison methods (not operator)
    def eq(self, other: Any) -> Boolean:
        return eq(self, other)

    def ne(self, other: Any) -> Boolean:
        return ne(self, other)

    # Comparison magic methods
    def __lt__(self, other: Any) -> Boolean:
        """Implements less than comparison using the < operator."""
        return lt(self, other)

    def __le__(self, other: Any) -> Boolean:
        """Implements less than or equal comparison using the <= operator."""
        return le(self, other)

    def __gt__(self, other: Any) -> Boolean:
        """Implements greater than comparison using the > operator."""
        return gt(self, other)

    def __ge__(self, other: Any) -> Boolean:
        """Implements greater than or equal comparison using the >= operator."""
        return ge(self, other)

    # Emulating container types
    def __getitem__(self, key: Any) -> Symbol:
        """Defines behavior for when an item is accessed,
        using the notation self[key]."""
        return getitem(self, key)

    # Emulating attribute
    def __getattr__(self, key: str) -> Symbol:
        """Defines behavior for when an item is accessed,
        using the notation self.key"""
        if key.startswith("__"):
            raise AttributeError(key)
        return symgetattr(self, key)

    # Normal arithmetic operators
    def __add__(self, other: Any) -> Symbol:
        """Implements addition."""
        return add(self, other)

    def __sub__(self, other: Any) -> Symbol:
        """Implements subtraction."""
        return sub(self, other)

    def __mul__(self, other: Any) -> Symbol:
        """Implements multiplication."""
        return mul(self, other)

    def __matmul__(self, other: Any) -> Symbol:
        """Implements multiplication."""
        return matmul(self, other)

    def __truediv__(self, other: Any) -> Symbol:
        """Implements true division."""
        return truediv(self, other)

    def __floordiv__(self, other: Any) -> Symbol:
        """Implements integer division using the // operator."""
        return floordiv(self, other)

    def __mod__(self, other: Any) -> Symbol:
        """Implements modulo using the % operator."""
        return mod(self, other)

    def __pow__(self, other: Any, modulo: Any = None) -> Symbol:
        """Implements behavior for exponents using the ** operator."""
        if modulo is None:
            return pow(self, other)
        else:
            return pow3(self, other, modulo)

    def __lshift__(self, other: Any) -> Symbol:
        """Implements left bitwise shift using the << operator."""
        return lshift(self, other)

    def __rshift__(self, other: Any) -> Symbol:
        """Implements right bitwise shift using the >> operator."""
        return rshift(self, other)

    def __and__(self, other: Any) -> Symbol:
        """Implements bitwise and using the & operator."""
        return and_(self, other)

    def __or__(self, other: Any) -> Symbol:
        """Implements bitwise or using the | operator."""
        return or_(self, other)

    def __xor__(self, other: Any) -> Symbol:
        """Implements bitwise xor using the ^ operator."""
        return xor(self, other)

    # Reflected arithmetic operators
    def __radd__(self, other: Any) -> Symbol:
        """Implements reflected addition."""
        return add(other, self)

    def __rsub__(self, other: Any) -> Symbol:
        """Implements reflected subtraction."""
        return sub(other, self)

    def __rmul__(self, other: Any) -> Symbol:
        """Implements reflected multiplication."""
        return mul(other, self)

    def __rmatmul__(self, other: Any) -> Symbol:
        """Implements reflected multiplication."""
        return matmul(other, self)

    def __rtruediv__(self, other: Any) -> Symbol:
        """Implements reflected true division."""
        return truediv(other, self)

    def __rfloordiv__(self, other: Any) -> Symbol:
        """Implements reflected integer division using the // operator."""
        return floordiv(other, self)

    def __rmod__(self, other: Any) -> Symbol:
        """Implements reflected modulo using the % operator."""
        return mod(other, self)

    def __rpow__(self, other: Any) -> Symbol:
        """Implements behavior for reflected exponents using the ** operator."""
        return pow(other, self)

    def __rlshift__(self, other: Any) -> Symbol:
        """Implements reflected left bitwise shift using the << operator."""
        return lshift(other, self)

    def __rrshift__(self, other: Any) -> Symbol:
        """Implements reflected right bitwise shift using the >> operator."""
        return rshift(other, self)

    def __rand__(self, other: Any) -> Symbol:
        """Implements reflected bitwise and using the & operator."""
        return and_(other, self)

    def __ror__(self, other: Any) -> Symbol:
        """Implements reflected bitwise or using the | operator."""
        return or_(other, self)

    def __rxor__(self, other: Any) -> Symbol:
        """Implements reflected bitwise xor using the ^ operator."""
        return xor(other, self)

    # Unary operators and functions
    def __neg__(self) -> Symbol:
        """Implements behavior for negation (e.g. -some_object)"""
        return neg(self)

    def __pos__(self) -> Symbol:
        """Implements behavior for unary positive (e.g. +some_object)"""
        return pos(self)

    def __invert__(self) -> Symbol:
        """Implements behavior for inversion using the ~ operator."""
        return invert(self)


CompOp = BinaryOperator[Any, Any, Boolean]
BinOp = BinaryOperator[Any, Any, Symbol]
BinFun = BinaryOperator[Any, Any, Symbol]
UnOp = UnaryOperator[Symbol, Symbol]

# Comparison methods (not operator)
eq = CompOp("eq", "symbol", precedence=-5, fmt="{} == {}", output_type=Boolean)
ne = CompOp("ne", "symbol", precedence=-5, fmt="{} != {}", output_type=Boolean)

# Comparison
lt = CompOp("lt", "symbol", precedence=-5, fmt="{} < {}", output_type=Boolean)
le = CompOp("le", "symbol", precedence=-5, fmt="{} <= {}", output_type=Boolean)
gt = CompOp("gt", "symbol", precedence=-5, fmt="{} > {}", output_type=Boolean)
ge = CompOp("ge", "symbol", precedence=-5, fmt="{} >= {}", output_type=Boolean)

# Emulating container types
getitem = BinOp("getitem", "symbol", precedence=5, fmt="{}[{}]", output_type=Symbol)

# Emulating attribute
symgetattr = BinOp(
    "symgetattr", "symbol", precedence=5, fmt="{}.{}", output_type=Symbol
)

# Emulating numeric types
add = BinOp("add", "symbol", precedence=0, fmt="{} + {}", output_type=Symbol)
sub = BinOp("sub", "symbol", precedence=0, fmt="{} - {}", output_type=Symbol)
mul = BinOp("mul", "symbol", precedence=1, fmt="{} * {}", output_type=Symbol)
matmul = BinOp("matmul", "symbol", precedence=1, fmt="{} @ {}", output_type=Symbol)
truediv = BinOp("truediv", "symbol", precedence=1, fmt="{} / {}", output_type=Symbol)
floordiv = BinOp("floordiv", "symbol", precedence=1, fmt="{} // {}", output_type=Symbol)
mod = BinOp("mod", "symbol", precedence=1, fmt="{} % {}", output_type=Symbol)
pow = BinOp("pow", "symbol", precedence=3, fmt="{} ** {}", output_type=Symbol)
pow3 = Function3[Symbol, Symbol]("pow3", "symbol", output_type=Symbol)
lshift = BinOp("lshift", "symbol", precedence=-1, fmt="{} << {}", output_type=Symbol)
rshift = BinOp("rshift", "symbol", precedence=-1, fmt="{} >> {}", output_type=Symbol)
and_ = BinOp("and_", "symbol", precedence=-2, fmt="{} & {}", output_type=Symbol)
xor = BinOp("xor", "symbol", precedence=-3, fmt="{} ^ {}", output_type=Symbol)
or_ = BinOp("or_", "symbol", precedence=-4, fmt="{} | {}", output_type=Symbol)

# Unary operators
neg = UnOp("neg", "symbol", precedence=2, fmt="-{}", output_type=Symbol)
pos = UnOp("pos", "symbol", precedence=2, fmt="+{}", output_type=Symbol)
invert = UnOp("invert", "symbol", precedence=2, fmt="~{}", output_type=Symbol)



"""
symbolite.impl.libpythoncode.symbol
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Code-emitting counterparts for ``symbolite.abstract.symbol``.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

from symbolite.abstract import symbol as abstract_symbol

from ._codeexpr import CodeExpr, as_function, as_operator

eq = as_operator(abstract_symbol.eq)
ne = as_operator(abstract_symbol.ne)
lt = as_operator(abstract_symbol.lt)
le = as_operator(abstract_symbol.le)
gt = as_operator(abstract_symbol.gt)
ge = as_operator(abstract_symbol.ge)

getitem = as_operator(abstract_symbol.getitem)
symgetattr = as_operator(abstract_symbol.symgetattr)

add = as_operator(abstract_symbol.add)
sub = as_operator(abstract_symbol.sub)
mul = as_operator(abstract_symbol.mul)
matmul = as_operator(abstract_symbol.matmul)
truediv = as_operator(abstract_symbol.truediv)
floordiv = as_operator(abstract_symbol.floordiv)
mod = as_operator(abstract_symbol.mod)
pow = as_operator(abstract_symbol.pow)
pow3 = as_function(abstract_symbol.pow3)
lshift = as_operator(abstract_symbol.lshift)
rshift = as_operator(abstract_symbol.rshift)
and_ = as_operator(abstract_symbol.and_)
xor = as_operator(abstract_symbol.xor)
or_ = as_operator(abstract_symbol.or_)

neg = as_operator(abstract_symbol.neg)
pos = as_operator(abstract_symbol.pos)
invert = as_operator(abstract_symbol.invert)


def Symbol(name: str) -> CodeExpr:
    return CodeExpr(name)


__all__ = [
    "Symbol",
    "eq",
    "ne",
    "lt",
    "le",
    "gt",
    "ge",
    "getitem",
    "symgetattr",
    "add",
    "sub",
    "mul",
    "matmul",
    "truediv",
    "floordiv",
    "mod",
    "pow",
    "pow3",
    "lshift",
    "rshift",
    "and_",
    "xor",
    "or_",
    "neg",
    "pos",
    "invert",
]



"""
symbolite.abstract.boolean
~~~~~~~~~~~~~~~~~~~~~~~~~~

Objects and functions for boolean operations.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

from typing import Self

from ..core import Value
from ..core.function import BinaryOperator


class Boolean(Value[bool]):
    """Boolean symbolic value.

    See Symbol and Value for more information.
    """

    def __and__(self, other: Self) -> Boolean:
        """Implements bitwise and using the & operator."""
        return and_(self, other)

    def __or__(self, other: Self) -> Boolean:
        """Implements bitwise or using the | operator."""
        return or_(self, other)

    def __xor__(self, other: Self) -> Boolean:
        """Implements bitwise xor using the ^ operator."""
        return xor(self, other)

    def __rand__(self, other: Self) -> Boolean:
        """Implements reflected bitwise and using the & operator."""
        return and_(other, self)

    def __ror__(self, other: Self) -> Boolean:
        """Implements reflected bitwise or using the | operator."""
        return or_(other, self)

    def __rxor__(self, other: Self) -> Boolean:
        """Implements reflected bitwise xor using the ^ operator."""
        return xor(other, self)


BinOp = BinaryOperator[Boolean, Boolean, Boolean]

and_ = BinOp("and_", "boolean", precedence=-2, fmt="{} & {}", output_type=Boolean)
xor = BinOp("xor", "boolean", precedence=-3, fmt="{} ^ {}", output_type=Boolean)
or_ = BinOp("or_", "boolean", precedence=-4, fmt="{} | {}", output_type=Boolean)

del BinOp



"""
symbolite.impl.libpythoncode.real
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Code-emitting counterparts for ``symbolite.abstract.real``.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

from symbolite.abstract import real as abstract_real

from ._codeexpr import (
    CodeExpr,
    as_function,
    as_named_value,
    as_operator,
    make_function,
)

eq = as_operator(abstract_real.eq)
ne = as_operator(abstract_real.ne)
lt = as_operator(abstract_real.lt)
le = as_operator(abstract_real.le)
gt = as_operator(abstract_real.gt)
ge = as_operator(abstract_real.ge)


def Real(name: str) -> CodeExpr:
    return CodeExpr(name)


add = as_operator(abstract_real.add)
sub = as_operator(abstract_real.sub)
mul = as_operator(abstract_real.mul)
truediv = as_operator(abstract_real.truediv)
floordiv = as_operator(abstract_real.floordiv)
mod = as_operator(abstract_real.mod)
lshift = as_operator(abstract_real.lshift)
rshift = as_operator(abstract_real.rshift)
and_ = as_operator(abstract_real.and_)
xor = as_operator(abstract_real.xor)
or_ = as_operator(abstract_real.or_)

neg = as_operator(abstract_real.neg)
pos = as_operator(abstract_real.pos)
invert = as_operator(abstract_real.invert)

abs = as_function(abstract_real.abs)
acos = as_function(abstract_real.acos)
acosh = as_function(abstract_real.acosh)
asin = as_function(abstract_real.asin)
asinh = as_function(abstract_real.asinh)
atan = as_function(abstract_real.atan)
atan2 = as_function(abstract_real.atan2)
atanh = as_function(abstract_real.atanh)
ceil = as_function(abstract_real.ceil)
comb = as_function(abstract_real.comb)
copysign = as_function(abstract_real.copysign)
cos = as_function(abstract_real.cos)
cosh = as_function(abstract_real.cosh)
degrees = as_function(abstract_real.degrees)
erf = as_function(abstract_real.erf)
erfc = as_function(abstract_real.erfc)
exp = as_function(abstract_real.exp)
expm1 = as_function(abstract_real.expm1)
fabs = as_function(abstract_real.fabs)
factorial = as_function(abstract_real.factorial)
floor = as_function(abstract_real.floor)
fmod = as_function(abstract_real.fmod)
frexp = as_function(abstract_real.frexp)
gamma = as_function(abstract_real.gamma)
hypot = as_function(abstract_real.hypot)
isfinite = as_function(abstract_real.isfinite)
isinf = as_function(abstract_real.isinf)
isnan = as_function(abstract_real.isnan)
isqrt = as_function(abstract_real.isqrt)
ldexp = as_function(abstract_real.ldexp)
lgamma = as_function(abstract_real.lgamma)
log = as_function(abstract_real.log)
log10 = as_function(abstract_real.log10)
log1p = as_function(abstract_real.log1p)
log2 = as_function(abstract_real.log2)
modf = as_function(abstract_real.modf)
nextafter = as_function(abstract_real.nextafter)
radians = as_function(abstract_real.radians)
remainder = as_function(abstract_real.remainder)
sin = as_function(abstract_real.sin)
sinh = as_function(abstract_real.sinh)
sqrt = as_function(abstract_real.sqrt)
tan = as_function(abstract_real.tan)
tanh = as_function(abstract_real.tanh)
trunc = as_function(abstract_real.trunc)
ulp = as_function(abstract_real.ulp)

pow3 = as_function(abstract_real.pow3_op)

_pow_call = make_function("real.pow")
_pow_operator = as_operator(abstract_real.pow_op)


def pow(*args: object, **kwargs: object):
    if len(args) == 2 and not kwargs:
        return _pow_operator(*args)
    return _pow_call(*args, **kwargs)


e = as_named_value(abstract_real.e)
inf = as_named_value(abstract_real.inf)
pi = as_named_value(abstract_real.pi)
nan = as_named_value(abstract_real.nan)
tau = as_named_value(abstract_real.tau)

__all__ = [
    "Real",
    "eq",
    "ne",
    "lt",
    "le",
    "gt",
    "ge",
    "add",
    "sub",
    "mul",
    "truediv",
    "floordiv",
    "mod",
    "lshift",
    "rshift",
    "and_",
    "xor",
    "or_",
    "neg",
    "pos",
    "invert",
    "abs",
    "acos",
    "acosh",
    "asin",
    "asinh",
    "atan",
    "atan2",
    "atanh",
    "ceil",
    "comb",
    "copysign",
    "cos",
    "cosh",
    "degrees",
    "erf",
    "erfc",
    "exp",
    "expm1",
    "fabs",
    "factorial",
    "floor",
    "fmod",
    "frexp",
    "gamma",
    "hypot",
    "isfinite",
    "isinf",
    "isnan",
    "isqrt",
    "ldexp",
    "lgamma",
    "log",
    "log10",
    "log1p",
    "log2",
    "modf",
    "nextafter",
    "radians",
    "remainder",
    "sin",
    "sinh",
    "sqrt",
    "tan",
    "tanh",
    "trunc",
    "ulp",
    "pow",
    "pow3",
    "e",
    "inf",
    "pi",
    "nan",
    "tau",
]



"""
symbolite.abstract.real
~~~~~~~~~~~~~~~~~~~~~~~

Objects and functions for real number operations.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

from typing import Any

from ..core import Value
from ..core.function import (
    BinaryFunction,
    BinaryOperator,
    Function3,
    UnaryFunction,
    UnaryOperator,
)
from .boolean import Boolean

NumberT = int | float


class Real(Value[NumberT]):
    """Real number symbolic value.

    See Symbol and Value for more information.
    """

    def eq(self, other: Any) -> Boolean:
        return eq(self, other)

    def ne(self, other: Any) -> Boolean:
        return ne(self, other)

    def __lt__(self, other: Any) -> Boolean:
        return lt(self, other)

    def __le__(self, other: Any) -> Boolean:
        return le(self, other)

    def __gt__(self, other: Any) -> Boolean:
        return gt(self, other)

    def __ge__(self, other: Any) -> Boolean:
        return ge(self, other)

    # Normal arithmetic operators
    def __add__(self, other: Any) -> Real:
        """Implements addition."""
        return add(self, other)

    def __sub__(self, other: Any) -> Real:
        """Implements subtraction."""
        return sub(self, other)

    def __mul__(self, other: Any) -> Real:
        """Implements multiplication."""
        return mul(self, other)

    def __truediv__(self, other: Any) -> Real:
        """Implements true division."""
        return truediv(self, other)

    def __floordiv__(self, other: Any) -> Real:
        """Implements integer division using the // operator."""
        return floordiv(self, other)

    def __mod__(self, other: Any) -> Real:
        """Implements modulo using the % operator."""
        return mod(self, other)

    def __pow__(self, other: Any, modulo: Any = None) -> Real:
        """Implements behavior for exponents using the ** operator."""
        if modulo is None:
            return pow_op(self, other)
        else:
            return pow3_op(self, other, modulo)

    def __lshift__(self, other: Any) -> Real:
        """Implements left bitwise shift using the << operator."""
        return lshift(self, other)

    def __rshift__(self, other: Any) -> Real:
        """Implements right bitwise shift using the >> operator."""
        return rshift(self, other)

    def __and__(self, other: Any) -> Real:
        """Implements bitwise and using the & operator."""
        return and_(self, other)

    def __or__(self, other: Any) -> Real:
        """Implements bitwise or using the | operator."""
        return or_(self, other)

    def __xor__(self, other: Any) -> Real:
        """Implements bitwise xor using the ^ operator."""
        return xor(self, other)

    # Reflected arithmetic operators
    def __radd__(self, other: Any) -> Real:
        """Implements reflected addition."""
        return add(other, self)

    def __rsub__(self, other: Any) -> Real:
        """Implements reflected subtraction."""
        return sub(other, self)

    def __rmul__(self, other: Any) -> Real:
        """Implements reflected multiplication."""
        return mul(other, self)

    def __rtruediv__(self, other: Any) -> Real:
        """Implements reflected true division."""
        return truediv(other, self)

    def __rfloordiv__(self, other: Any) -> Real:
        """Implements reflected integer division using the // operator."""
        return floordiv(other, self)

    def __rmod__(self, other: Any) -> Real:
        """Implements reflected modulo using the % operator."""
        return mod(other, self)

    def __rpow__(self, other: Any) -> Real:
        """Implements behavior for reflected exponents using the ** operator."""
        return pow_op(other, self)

    def __rlshift__(self, other: Any) -> Real:
        """Implements reflected left bitwise shift using the << operator."""
        return lshift(other, self)

    def __rrshift__(self, other: Any) -> Real:
        """Implements reflected right bitwise shift using the >> operator."""
        return rshift(other, self)

    def __rand__(self, other: Any) -> Real:
        """Implements reflected bitwise and using the & operator."""
        return and_(other, self)

    def __ror__(self, other: Any) -> Real:
        """Implements reflected bitwise or using the | operator."""
        return or_(other, self)

    def __rxor__(self, other: Any) -> Real:
        """Implements reflected bitwise xor using the ^ operator."""
        return xor(other, self)

    # Unary operators and functions
    def __neg__(self) -> Real:
        """Implements behavior for negation (e.g. -some_object)"""
        return neg(self)

    def __pos__(self) -> Real:
        """Implements behavior for unary positive (e.g. +some_object)"""
        return pos(self)

    def __invert__(self) -> Real:
        """Implements behavior for inversion using the ~ operator."""
        return invert(self)


CompOp = BinaryOperator[Real | NumberT, Real | NumberT, Boolean]

# Comparison methods (not operator)
eq = CompOp("eq", "real", precedence=-5, fmt="{} == {}", output_type=Boolean)
ne = CompOp("ne", "real", precedence=-5, fmt="{} != {}", output_type=Boolean)

# Comparison
lt = CompOp("lt", "real", precedence=-5, fmt="{} < {}", output_type=Boolean)
le = CompOp("le", "real", precedence=-5, fmt="{} <= {}", output_type=Boolean)
gt = CompOp("gt", "real", precedence=-5, fmt="{} > {}", output_type=Boolean)
ge = CompOp("ge", "real", precedence=-5, fmt="{} >= {}", output_type=Boolean)

BinOp = BinaryOperator[Real | NumberT, Real | NumberT, Real]

# Emulating numeric types
add = BinOp("add", "real", precedence=0, fmt="{} + {}", output_type=Real)
sub = BinOp("sub", "real", precedence=0, fmt="{} - {}", output_type=Real)
mul = BinOp("mul", "real", precedence=1, fmt="{} * {}", output_type=Real)
truediv = BinOp("truediv", "real", precedence=1, fmt="{} / {}", output_type=Real)
floordiv = BinOp("floordiv", "real", precedence=1, fmt="{} // {}", output_type=Real)
mod = BinOp("mod", "real", precedence=1, fmt="{} % {}", output_type=Real)
pow_op = BinOp("pow", "real", precedence=3, fmt="{} ** {}", output_type=Real)
pow3_op = Function3[Real, Real]("pow3", "real", output_type=Real)
lshift = BinOp("lshift", "real", precedence=-1, fmt="{} << {}", output_type=Real)
rshift = BinOp("rshift", "real", precedence=-1, fmt="{} >> {}", output_type=Real)
and_ = BinOp("and_", "real", precedence=-2, fmt="{} & {}", output_type=Real)
xor = BinOp("xor", "real", precedence=-3, fmt="{} ^ {}", output_type=Real)
or_ = BinOp("or_", "real", precedence=-4, fmt="{} | {}", output_type=Real)

UnOp = UnaryOperator[Real, Real]

# Unary operators
neg = UnOp("neg", "real", precedence=2, fmt="-{}", output_type=Real)
pos = UnOp("pos", "real", precedence=2, fmt="+{}", output_type=Real)
invert = UnOp("invert", "real", precedence=2, fmt="~{}", output_type=Real)

UnFun = UnaryFunction[Real | NumberT, Real]
BinFun = BinaryFunction[Real | NumberT, Real]

# "gcd": None,  # 1 to ---
# "hypot": None,  # 1 to ---
# "isclose": None,  # 2, 3, 4
# "lcm": None,  # 1 to ---
# "perm": None,  # 1 or 2
# "log": None,  # 1 or 2 is used as log(x, e)

abs = UnFun("abs", "real", output_type=Real)
acos = UnFun("acos", "real", output_type=Real)
acosh = UnFun("acosh", "real", output_type=Real)
asin = UnFun("asin", "real", output_type=Real)
asinh = UnFun("asinh", "real", output_type=Real)
atan = UnFun("atan", "real", output_type=Real)
atan2 = BinFun("atan2", "real", output_type=Real)
atanh = UnFun("atanh", "real", output_type=Real)
ceil = UnFun("ceil", "real", output_type=Real)
comb = UnFun("comb", "real", output_type=Real)
copysign = UnFun("copysign", "real", output_type=Real)
cos = UnFun("cos", "real", output_type=Real)
cosh = UnFun("cosh", "real", output_type=Real)
degrees = UnFun("degrees", "real", output_type=Real)
erf = UnFun("erf", "real", output_type=Real)
erfc = UnFun("erfc", "real", output_type=Real)
exp = UnFun("exp", "real", output_type=Real)
expm1 = UnFun("expm1", "real", output_type=Real)
fabs = UnFun("fabs", "real", output_type=Real)
factorial = UnFun("factorial", "real", output_type=Real)
floor = UnFun("floor", "real", output_type=Real)
fmod = UnFun("fmod", "real", output_type=Real)
frexp = UnFun("frexp", "real", output_type=Real)
gamma = UnFun("gamma", "real", output_type=Real)
hypot = UnFun("gamma", "real", output_type=Real)
isfinite = UnFun("isfinite", "real", output_type=Real)
isinf = UnFun("isinf", "real", output_type=Real)
isnan = UnFun("isnan", "real", output_type=Real)
isqrt = UnFun("isqrt", "real", output_type=Real)
ldexp = BinFun("ldexp", "real", output_type=Real)
lgamma = UnFun("lgamma", "real", output_type=Real)
log = UnFun("log", "real", output_type=Real)
log10 = UnFun("log10", "real", output_type=Real)
log1p = UnFun("log1p", "real", output_type=Real)
log2 = UnFun("log2", "real", output_type=Real)
modf = UnFun("modf", "real", output_type=Real)
nextafter = UnFun("nextafter", "real", output_type=Real)
pow = BinFun("pow", "real", output_type=Real)
radians = UnFun("radians", "real", output_type=Real)
remainder = BinFun("remainder", "real", output_type=Real)
sin = UnFun("sin", "real", output_type=Real)
sinh = UnFun("sinh", "real", output_type=Real)
sqrt = UnFun("sqrt", "real", output_type=Real)
tan = UnFun("tan", "real", output_type=Real)
tanh = UnFun("tanh", "real", output_type=Real)
tan = UnFun("tan", "real", output_type=Real)
trunc = UnFun("trunc", "real", output_type=Real)
ulp = UnFun("ulp", "real", output_type=Real)

e = Real("real.e")
inf = Real("real.inf")
pi = Real("real.pi")
nan = Real("real.nan")
tau = Real("rea.tau")

del (
    BinaryFunction,
    BinaryOperator,
    UnaryFunction,
    UnaryOperator,
    CompOp,
    BinOp,
    UnOp,
    BinFun,
    UnFun,
)
